# Copyright 2017 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kstream.includes.windows cimport UCHAR, ULONG, wchar_t

cdef struct PROCESS_INFO:
    # process identifier
    ULONG pid
    # process parent identifier
    ULONG ppid
    # name of the image file
    wchar_t* name


# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import patch

import elasticsearch
import pytest

from fibratus.errors import InvalidPayloadError
from fibratus.output.elasticsearch import ElasticsearchOutput


@pytest.fixture(scope='module')
def elasticsearch_adapter():
    config = {
        'hosts': [
            'localhost:9200',
            'rabbitstack:9200'
        ],
        'index': 'kernelstream',
        'document': 'threads'
    }
    return ElasticsearchOutput(**config)


@pytest.fixture(scope='module')
def elasticsearch_bulk_adapter():
    config = {
        'hosts': [
            'localhost:9200',
            'rabbitstack:9200'
        ],
        'index': 'kernelstream',
        'document': 'threads',
        'bulk': True,
        'username': 'elastic',
        'password': 'changeme'
    }
    return ElasticsearchOutput(**config)


class TestElasticsearchOutput(object):

    def test_init(self, elasticsearch_adapter):
        assert isinstance(elasticsearch_adapter.hosts, list)
        assert len(elasticsearch_adapter.hosts) > 0
        assert {'host': 'localhost', 'port': 9200} in elasticsearch_adapter.hosts
        assert elasticsearch_adapter.index_name == 'kernelstream'
        assert elasticsearch_adapter.document_type == 'threads'
        assert not elasticsearch_adapter.bulk

    def test_emit(self, elasticsearch_adapter):
        body = {'kevent_type': 'CreateProcess', 'params': {'name': 'smss.exe'}}
        assert elasticsearch_adapter._elasticsearch is None
        with patch('elasticsearch.Elasticsearch', spec_set=elasticsearch.Elasticsearch) as es_client_mock:
            elasticsearch_adapter.emit(body)
            es_client_mock.assert_called_with([{'host': 'localhost', 'port': 9200},
                                               {'host': 'rabbitstack', 'port': 9200}], use_ssl=False)
            elasticsearch_adapter._elasticsearch.index.assert_called_with('kernelstream', 'threads', body=body)

    @patch('elasticsearch.Elasticsearch', spec_set=elasticsearch.Elasticsearch)
    def test_emit_invalid_payload(self, es_client_mock, elasticsearch_adapter):
        body = ['CreateProcess', 'TerminateProcess']
        with pytest.raises(InvalidPayloadError) as e:
            elasticsearch_adapter.emit(body)
        assert "invalid payload for document. dict expected but <class 'list'> found" == str(e.value)
        assert es_client_mock.index.assert_not_called()

    @patch('elasticsearch.Elasticsearch', spec_set=elasticsearch.Elasticsearch)
    @patch('elasticsearch.helpers.bulk')
    def test_emit_bulk(self, es_bulk_mock, es_client_mock, elasticsearch_bulk_adapter):
        body = [{'kevent_type': 'CreateProcess', 'params': {'name': 'smss.exe'}},
                {'kevent_type': 'TerminateProcess', 'params': {'name': 'smss.exe'}}]
        elasticsearch_bulk_adapter.emit(body)
        es_client_mock.assert_called_with([{'host': 'localhost', 'port': 9200},
                                           {'host': 'rabbitstack', 'port': 9200}], use_ssl=False,
                                          http_auth=('elastic', 'changeme',))
        expected_body = [{'_index': 'kernelstream', '_type': 'threads', '_source':
                                    {'kevent_type': 'CreateProcess', 'params': {'name': 'smss.exe'}}},
                         {'_index': 'kernelstream', '_type': 'threads', '_source':
                                    {'kevent_type': 'TerminateProcess', 'params': {'name': 'smss.exe'}}}]
        es_bulk_mock.assert_called_once_with(elasticsearch_bulk_adapter._elasticsearch, expected_body)

    @patch('elasticsearch.Elasticsearch', spec_set=elasticsearch.Elasticsearch)
    @patch('elasticsearch.helpers.bulk')
    def test_emit_bulk_invalid_payload(self, es_bulk_mock, es_client_mock, elasticsearch_bulk_adapter):
        body = ({'kevent_type': 'CreateProcess', 'params': {'name': 'smss.exe'}},
                {'kevent_type': 'TerminateProcess', 'params': {'name': 'smss.exe'}},)
        with pytest.raises(InvalidPayloadError) as e:
            elasticsearch_bulk_adapter.emit(body)
        assert "invalid payload for bulk indexing. list expected but <class 'tuple'> found" == str(e.value)
        assert es_bulk_mock.assert_not_called()



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os


class DllRepository(object):

    def __init__(self, kevent):
        self.dlls = {}
        self._kevent = kevent

    def register_dll(self, kdll):
        """Registers a loaded image.

        Registers an image when
        the latter is loaded into the address space
        of the process.

        Parameters
        ----------

        kdll: dict
            Image load event payload as forwarded
            from the kernel event stream collector
        """
        pid = kdll.process_id
        path = kdll.file_name
        image = os.path.basename(path)
        size = kdll.image_size
        checksum = kdll.image_checksum
        base = kdll.image_base

        self._kevent.pid = pid

        dll = Dll(pid, path,
                  image,
                  size,
                  checksum,
                  base)

        if pid in self.dlls:
            # append a new image to
            # the associated process
            self.dlls[pid].append(dll)
        else:
            self.dlls[pid] = [dll]

        self._kevent.params = dict(image=image,
                                   pid=pid,
                                   path=path,
                                   size=dll.size,
                                   checksum=checksum,
                                   base=hex(base))

    def unregister_dll(self, kdll):
        """Unregisters a loaded image.

        Removes the loaded image from
        the repository for a given process.

        Parameters
        ----------

        kdll: dict
            Image unload event payload as forwarded
            from the kernel event stream collector
        """
        pid = kdll.process_id
        path = kdll.file_name
        image = os.path.basename(path)
        size = kdll.image_size / 1024
        checksum = kdll.image_checksum
        base = kdll.image_base

        self._kevent.pid = pid

        if pid in self.dlls:
            dlls = self.dlls[pid]
            for dll in dlls:
                if dll.image == image:
                    dlls.remove(dll)
        self._kevent.params = dict(image=image,
                                   pid=pid,
                                   path=path,
                                   size=size,
                                   checksum=checksum,
                                   base=hex(base))

    def dlls_for_process(self, pid):
        return self.dlls[pid] if pid in self.dlls else []


class Dll(object):

    def __init__(self, pid, path, image, size, checksum, base):
        self._pid = pid
        self._path = path
        self._size = size
        self._checksum = checksum
        self._base = base
        self._image = image

    @property
    def pid(self):
        return self._pid

    @property
    def path(self):
        return self._path

    @property
    def image(self):
        return self._image

    @property
    def size(self):
        return self._size

    @property
    def base(self):
        return hex(self._base)

    @property
    def checksum(self):
        return self._checksum



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import smtplib

from fibratus.output.base import Output


class SmtpOutput(Output):

    def __init__(self, **kwargs):
        """Constructs a new instance of the SMTP outbound adapter.

        Parameters
        ----------

        kwargs: dict
            SMTP server and account configuration
        """
        Output.__init__(self)
        self._host = kwargs.pop('host', None)
        self._port = kwargs.pop('port', 587)
        self._from = kwargs.pop('from', None)
        self._to = kwargs.pop('to', [])
        self._password = kwargs.pop('password', None) or \
            os.environ.get('SMTP_PASS')
        self._smtp = None

    def emit(self, body, **kwargs):
        if not self._smtp:
            self._smtp = smtplib.SMTP(self._host, self._port)
        self._smtp.ehlo()
        self._smtp.starttls()
        self._smtp.ehlo()
        subject = kwargs.pop('subject', '')
        message = self._compose_message(subject, body)
        # try to authenticate with the server
        # before attempting to send the message
        try:
            self._smtp.login(self._from, self._password)
            self._smtp.sendmail(self._from, self._to, message)
        except smtplib.SMTPAuthenticationError:
            self.logger.error('Invalid SMTP credentials for %s account'
                              % self._from)
        finally:
            self._smtp.quit()

    def _compose_message(self, subject, body):
        return """From: %s\r\nTo: %s\r\nSubject: %s\r\n\

                    %s
                """ % (self._from, ", ".join(self._to),
                       subject, body)

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def sender(self):
        return self._from

    @property
    def to(self):
        return self._to



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Shows the top TCP / UDP incoming packets.
"""

import collections

connections = collections.Counter()


def on_init():
    set_filter('Recv')
    columns(["Source", "Count"])
    sort_by('Count')
    set_interval(1)
    title('Top incoming TCP/UDP packets')


def on_next_kevent(kevent):
    src = ['%s:%d' % (kevent.params.ip_dst, kevent.params.dport)]
    connections.update(src)


def on_interval():
    for ip, count in connections.items():
        add_row([ip, count])
    render_tabular()




# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import ctypes
import os
from ctypes.wintypes import MAX_PATH, DWORD
from enum import Enum

from fibratus.apidefs.process import open_thread, get_process_id_of_thread, \
    THREAD_QUERY_INFORMATION, open_process, PROCESS_QUERY_INFORMATION, PROCESS_VM_READ, \
    query_full_process_image_name
from fibratus.apidefs.sys import close_handle
from fibratus.common import NA


class ContextSwitchRegistry(object):
    """Keeps the state of the context switches ocurring on the system.

    Once the CPU scheduler selects a new thread to execute, the context switch
    registry tracks down a plethora of attributes like the new thread priority,
    the old thread state, the wait reason, etc. It also keeps a counter on how many
    context switches has been made for a particular thread and the logical cpu.

    """

    def __init__(self, thread_registry, kevent):
        self._css = {}
        self._thread_registry = thread_registry
        self._kevent = kevent

    def next_cswitch(self, cpu, ts, kcs, on_context_switch=None):
        """Parses the context switch kernel events.

        Parameters
        ----------
        cpu: int
            the logical cpu where the context switch occurs
        ts: str
            the timestamp of the context switch
        kcs: dict
            the context switch info as forwarded
            from the kstream collector
        on_context_switch: callable
            the callback to execute after the parsing stage

        """
        new_thread_id = int(kcs.new_thread_id, 16)
        old_thread_id = int(kcs.old_thread_id, 16)
        new_thread_wait_time = int(kcs.new_thread_wait_time, 16)
        thread_cs = (cpu, new_thread_id,)
        next_thread = self._thread_registry.get_thread(new_thread_id)
        prev_thread = self._thread_registry.get_thread(old_thread_id)

        next_pid = next_thread.pid if next_thread else None
        next_proc_name = next_thread.name if next_thread \
            else self._get_proc(new_thread_id)
        prev_proc_name = prev_thread.name if prev_thread \
            else self._get_proc(old_thread_id)

        if thread_cs in self._css:
            # if the thread has been previously scheduled
            # on the same logical cpu, we can update its
            # context switch info
            cs = self._css[thread_cs]
            cs.timestamp = ts
            cs.prev_thread = prev_proc_name or NA
            cs.next_thread_prio = kcs.new_thread_priority
            cs.next_thread_wait_time = new_thread_wait_time
            cs.prev_thread_prio = kcs.old_thread_priority
            cs.prev_thread_state = ContextSwitchRegistry._human_thread_state(kcs.old_thread_state)
            cs.prev_thread_wait_mode = ContextSwitchRegistry._human_wait_mode(kcs.old_thread_wait_mode)
            cs.prev_thread_wait_reason = ContextSwitchRegistry._human_wait_reason(kcs.old_thread_wait_reason)
            cs.increment_count()
        else:
            # the new thread has been scheduled
            # add it to the registry of context
            # switches
            cs = CSwitch(ts,
                         next_proc_name or NA,
                         prev_proc_name or NA,
                         kcs.new_thread_priority,
                         new_thread_wait_time,
                         kcs.old_thread_priority,
                         ContextSwitchRegistry._human_thread_state(kcs.old_thread_state),
                         ContextSwitchRegistry._human_wait_mode(kcs.old_thread_wait_mode),
                         ContextSwitchRegistry._human_wait_reason(kcs.old_thread_wait_reason))
            cs.increment_count()
            self._css[thread_cs] = cs

        if on_context_switch:
            if next_proc_name:
                on_context_switch(cpu, next_proc_name)
            else:
                on_context_switch(cpu, kcs.new_thread_id)

        self._kevent.tid = new_thread_id
        self._kevent.pid = next_pid
        params = {
            'next_proc_name': cs.next_proc_name,
            'prev_proc_name': cs.prev_proc_name,
            'cpu': cpu,
            'next_thread_id': new_thread_id,
            'prev_thread_id': old_thread_id,
            'next_thread_prio': cs.next_thread_prio,
            'prev_thread_prio': cs.prev_thread_prio,
            'prev_thread_state': cs.prev_thread_state.name if cs.prev_thread_state else NA,
            'next_thread_wait_time': cs.next_thread_wait_time,
            'prev_thread_wait_mode': cs.prev_thread_wait_mode.name if cs.prev_thread_wait_mode else NA,
            'prev_thread_wait_reason': cs.prev_thread_wait_reason.name if cs.prev_thread_wait_reason else NA
        }
        self._kevent.params = params

    def context_switches(self):
        """Returns a dictionary of context switches.
        """
        return self._css

    def _get_proc(self, thread_id):
        handle = open_thread(THREAD_QUERY_INFORMATION,
                             False,
                             thread_id)

        if handle:
            # if it was possible to get the process id
            # which is the parent of the thread, we can
            # try to get the process name from its pid
            pid = get_process_id_of_thread(handle)
            close_handle(handle)
            handle = open_process(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ,
                                  False,
                                  pid)
            if handle:
                exe = ctypes.create_unicode_buffer(MAX_PATH)
                status = query_full_process_image_name(handle, 0,
                                                       exe, DWORD(MAX_PATH))
                close_handle(handle)
                if status:
                    return os.path.basename(exe.value)

    @classmethod
    def _human_thread_state(cls, thread_state):
        if thread_state == ThreadState.INITIALIZED.value:
            return ThreadState.INITIALIZED
        elif thread_state == ThreadState.READY.value:
            return ThreadState.READY
        elif thread_state == ThreadState.RUNNING.value:
            return ThreadState.RUNNING
        elif thread_state == ThreadState.STANDBY.value:
            return ThreadState.STANDBY
        elif thread_state == ThreadState.TERMINATED.value:
            return ThreadState.TERMINATED
        elif thread_state == ThreadState.WAITING.value:
            return ThreadState.WAITING
        elif thread_state == ThreadState.TRANSITION.value:
            return ThreadState.TRANSITION
        elif thread_state == ThreadState.DEFERRED_READY.value:
            return ThreadState.DEFERRED_READY

    @classmethod
    def _human_wait_reason(cls, wait_reason):
        if wait_reason == WaitReason.EXECUTIVE.value or wait_reason == WaitReason.EXECUTIVE.value + 7:
            return WaitReason.EXECUTIVE
        elif wait_reason == WaitReason.FREE_PAGE.value or wait_reason == WaitReason.FREE_PAGE.value + 7:
            return WaitReason.FREE_PAGE
        elif wait_reason == WaitReason.PAGE_IN.value or wait_reason == WaitReason.PAGE_IN.value + 7:
            return WaitReason.PAGE_IN
        elif wait_reason == WaitReason.POOL_ALLOCATION.value or wait_reason == WaitReason.POOL_ALLOCATION.value + 7:
            return WaitReason.POOL_ALLOCATION
        elif wait_reason == WaitReason.DELAY_EXECUTION.value or wait_reason == WaitReason.DELAY_EXECUTION.value + 7:
            return WaitReason.DELAY_EXECUTION
        elif wait_reason == WaitReason.SUSPENDED.value or wait_reason == WaitReason.SUSPENDED.value + 7:
            return WaitReason.SUSPENDED
        elif wait_reason == WaitReason.USER_REQUEST or wait_reason == WaitReason.USER_REQUEST.value + 7:
            return WaitReason.USER_REQUEST
        elif wait_reason == WaitReason.EVENT_PAIR.value:
            return WaitReason.EVENT_PAIR
        elif wait_reason == WaitReason.QUEUE.value:
            return WaitReason.QUEUE
        elif wait_reason == WaitReason.LPC_RECEIVE.value:
            return WaitReason.LPC_RECEIVE
        elif wait_reason == WaitReason.LPC_REPLY.value:
            return WaitReason.LPC_REPLY
        elif wait_reason == WaitReason.VIRTUAL_MEMORY.value:
            return WaitReason.VIRTUAL_MEMORY
        elif wait_reason == WaitReason.PAGE_OUT.value:
            return WaitReason.PAGE_OUT
        elif wait_reason == WaitReason.RENDEZVOUS.value:
            return WaitReason.RENDEZVOUS
        elif wait_reason == WaitReason.KEYED_EVENT.value:
            return WaitReason.KEYED_EVENT
        elif wait_reason == WaitReason.TERMINATED.value:
            return WaitReason.TERMINATED
        elif wait_reason == WaitReason.PROCESS_IN_SWAP.value:
            return WaitReason.PROCESS_IN_SWAP
        elif wait_reason == WaitReason.CPU_WAIT_CONTROL.value:
            return WaitReason.CPU_WAIT_CONTROL
        elif wait_reason == WaitReason.CALLOUT_STACK.value:
            return WaitReason.CALLOUT_STACK
        elif wait_reason == WaitReason.KERNEL.value:
            return WaitReason.KERNEL
        elif wait_reason == WaitReason.RESOURCE.value:
            return WaitReason.RESOURCE
        elif wait_reason == WaitReason.PUSH_LOCK.value:
            return WaitReason.PUSH_LOCK
        elif wait_reason == WaitReason.MUTEX.value:
            return WaitReason.MUTEX
        elif wait_reason == WaitReason.QUANTUM_END.value:
            return WaitReason.QUANTUM_END
        elif wait_reason == WaitReason.DISPATCH_INT.value:
            return WaitReason.DISPATCH_INT
        elif wait_reason == WaitReason.PREEMPTED.value:
            return WaitReason.PREEMPTED
        elif wait_reason == WaitReason.YIELD_EXECUTION.value:
            return WaitReason.YIELD_EXECUTION
        elif wait_reason == WaitReason.FAST_MUTEX.value:
            return WaitReason.FAST_MUTEX
        elif wait_reason == WaitReason.GUARDED_MUTEX.value:
            return WaitReason.GUARDED_MUTEX
        elif wait_reason == WaitReason.RUNDOWN.value:
            return WaitReason.RUNDOWN
        elif wait_reason == WaitReason.MAXIMUM_WAIT_REASON.value:
            return WaitReason.MAXIMUM_WAIT_REASON

    @classmethod
    def _human_wait_mode(cls, wait_mode):
        if wait_mode == WaitMode.KERNEL.value:
            return WaitMode.KERNEL
        elif wait_mode == WaitMode.USER.value:
            return WaitMode.USER


class CSwitch(object):

    def __init__(self, ts, next_proc_name, prev_proc_name, next_thread_prio,
                 next_thread_wait_time, prev_thread_prio, prev_thread_state,
                 prev_thread_wait_mode,
                 prev_thread_wait_reason):
        """Context switch state info.

        Parameters
        ----------
        ts: str
            the timestamp of the context switch
        next_proc_name: str
            process name of the thread which is about to be scheduled
        prev_proc_name: str
            process name right before the context switch
        next_thread_prio: int
            the priority of the new thread
        next_thread_wait_time: int
            wait time for the new thread
        prev_thread_prio: int
            the priority of the old thread
        prev_thread_state: Enum
            state of the previous thread
        prev_thread_wait_mode: Enum
            the wait mode of the old thread
        prev_thread_wait_reason: Enum
            the wait reason of the previous thread

        """
        self._ts = ts
        self._next_proc_name = next_proc_name
        self._prev_proc_name = prev_proc_name
        self._next_thread_prio = next_thread_prio
        self._next_thread_wait_time = next_thread_wait_time
        self._prev_thread_prio = prev_thread_prio
        self._prev_thread_state = prev_thread_state
        self._prev_thread_wait_mode = prev_thread_wait_mode
        self._prev_thread_wait_reason = prev_thread_wait_reason
        self._count = 0

    @property
    def timestamp(self):
        return self._ts

    @timestamp.setter
    def timestamp(self, ts):
        self._ts = ts

    @property
    def next_proc_name(self):
        return self._next_proc_name

    @property
    def prev_proc_name(self):
        return self._prev_proc_name

    @property
    def next_thread_prio(self):
        return self._next_thread_prio

    @next_thread_prio.setter
    def next_thread_prio(self, next_thread_prio):
        self._next_thread_prio = next_thread_prio

    @property
    def next_thread_wait_time(self):
        return self._next_thread_wait_time

    @next_thread_wait_time.setter
    def next_thread_wait_time(self, next_thread_wait_time):
        self._next_thread_wait_time = next_thread_wait_time

    @property
    def prev_thread_prio(self):
        return self._prev_thread_prio

    @prev_thread_prio.setter
    def prev_thread_prio(self, prev_thread_prio):
        self._prev_thread_prio = prev_thread_prio

    @property
    def prev_thread_state(self):
        return self._prev_thread_state

    @prev_thread_state.setter
    def prev_thread_state(self, prev_thread_state):
        self._prev_thread_state = prev_thread_state

    @property
    def prev_thread_wait_mode(self):
        return self._prev_thread_wait_mode

    @prev_thread_wait_mode.setter
    def prev_thread_wait_mode(self, prev_thread_wait_mode):
        self._prev_thread_wait_mode = prev_thread_wait_mode

    @property
    def prev_thread_wait_reason(self):
        return self._prev_thread_wait_reason

    @prev_thread_wait_reason.setter
    def prev_thread_wait_reason(self, prev_thread_wait_reason):
        self._prev_thread_wait_reason = prev_thread_wait_reason

    @property
    def count(self):
        return self._count

    def increment_count(self):
        self._count += 1


class ThreadState(Enum):
    """Possible thread states.
    """
    INITIALIZED = 0
    READY = 1
    RUNNING = 2
    STANDBY = 3
    TERMINATED = 4
    WAITING = 5
    TRANSITION = 6
    DEFERRED_READY = 7


class WaitMode(Enum):
    KERNEL = 0
    USER = 1


class WaitReason(Enum):
    EXECUTIVE = 0
    FREE_PAGE = 1
    PAGE_IN = 2
    POOL_ALLOCATION = 3
    DELAY_EXECUTION = 4
    SUSPENDED = 5
    USER_REQUEST = 6
    EVENT_PAIR = 14
    QUEUE = 15
    LPC_RECEIVE = 16
    LPC_REPLY = 17
    VIRTUAL_MEMORY = 18
    PAGE_OUT = 19
    RENDEZVOUS = 20
    KEYED_EVENT = 21
    TERMINATED = 22
    PROCESS_IN_SWAP = 23
    CPU_WAIT_CONTROL = 24
    CALLOUT_STACK = 25
    KERNEL = 26
    RESOURCE = 27
    PUSH_LOCK = 28
    MUTEX = 29
    QUANTUM_END = 30
    DISPATCH_INT = 31
    PREEMPTED = 32
    YIELD_EXECUTION = 33
    FAST_MUTEX = 34
    GUARDED_MUTEX = 35
    RUNDOWN = 36
    MAXIMUM_WAIT_REASON = 37



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock
import os
import pytest

from fibratus.image_meta import ImageMetaRegistry


@pytest.fixture(scope='module')
def image_meta_registry():
    image_meta_registry = ImageMetaRegistry(imports=True, file_info=True)
    return image_meta_registry


@pytest.fixture(scope='module')
def image_path():
    return '%s\\notepad.exe' % os.environ['WINDIR']


class TestImageMetaRegistry(object):

    def test_add_image_meta(self, image_meta_registry, image_path):
        image_meta_registry.add_image_meta(image_path)
        image_meta = image_meta_registry.get_image_meta(image_path)
        assert image_meta

        assert image_meta.arch
        assert image_meta.timestamp
        assert image_meta.num_sections > 0

        assert len(image_meta.sections) > 0
        section_names = [se['name'] for se in image_meta.sections]

        assert '.text' in section_names

        assert len(image_meta.imports) > 0
        assert 'KERNEL32.dll' in image_meta.imports

        assert 'Microsoft Corporation' in image_meta.org
        assert image_meta.description
        assert image_meta.version
        assert 'Notepad' in image_meta.internal_name
        assert image_meta.copyright



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import patch
from fibratus.config import YamlConfig

import os

__CONFIG_PATH__ = os.path.join(os.path.dirname(__file__), '..', 'fixtures', 'fibratus.yml')
__SCHEMA_FILE__ = os.path.join(os.path.dirname(__file__), '..', 'fixtures', 'schema.yml')


class TestYamlConfig():

    def test_load_yaml(self):
        config = YamlConfig(__CONFIG_PATH__)
        config.load(False)
        assert config.yaml

    def test_load_validate(self):
        config = YamlConfig(__CONFIG_PATH__)
        config.default_schema_path = __SCHEMA_FILE__
        config.load()

    def test_load_yaml_not_found(self):
        with patch('sys.exit') as sys_exit:
            YamlConfig('C:\\fibratus.yml').load(False)
            sys_exit.assert_called_once()

    def test_outputs(self):
        config = YamlConfig(__CONFIG_PATH__)
        config.load(False)
        outputs = config.outputs
        assert outputs
        assert isinstance(outputs, list)
        assert len(outputs) > 0

    def test_enum_outputs(self):
        config = YamlConfig(__CONFIG_PATH__)
        config.load(False)
        output_names = ['amqp', 'smtp', 'console', 'elasticsearch', 'fs']
        outputs = config.outputs
        if outputs:
            for output in outputs:
                output_name = next(iter(list(output.keys())), None)
                assert output_name in output_names

    def test_image_skips(self):
        config = YamlConfig(__CONFIG_PATH__)
        config.load(False)
        image_skips = config.skips.images
        assert image_skips
        assert isinstance(image_skips, list)
        assert 'smss.exe' in image_skips

    def test_bindings(self):
        config = YamlConfig(__CONFIG_PATH__)
        config.load(False)
        bindings = config.bindings
        binding_names = ['yara']
        assert bindings
        for binding in bindings:
            assert isinstance(binding, dict)
            binding_name = next(iter(list(binding.keys())), None)
            assert binding_name in binding_names


# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import Mock

import pytest

from fibratus.kevent import KEvent
from fibratus.output.aggregator import OutputAggregator
from fibratus.output.amqp import AmqpOutput
from fibratus.output.console import ConsoleOutput


@pytest.fixture(scope='module')
def outputs():
    return dict(amqp=Mock(spec_set=AmqpOutput), console=Mock(spec_set=ConsoleOutput))


@pytest.fixture(scope='module')
def output_aggregator(outputs):
    return OutputAggregator(outputs)


@pytest.fixture(scope='module')
def kevent():
    kevent_mock = Mock(spec_set=KEvent)
    kevent_mock.get_thread.return_value = (343, 'svchost.exe')
    return kevent_mock


class TestOutputAggregator(object):

    def test_aggregate(self, output_aggregator, kevent, outputs):
        output_aggregator.aggregate(kevent)
        for _, output in outputs.items():
            assert output.emit.call_count == 1




# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from ctypes import Structure
from ctypes import c_void_p, c_ubyte, c_ushort, c_ulong, c_size_t, c_wchar_p
import re
import ctypes

# undefined ctypes wintypes
LPVOID = c_void_p
PVOID = c_void_p
UCHAR = c_ubyte
SIZE_T = c_size_t
LPTSTR = c_wchar_p

# status codes
STATUS_INFO_LENGTH_MISMATCH = 0xc0000004
STATUS_SUCCESS = 0

# error codes
ERROR_SUCCESS = 0x0
ERROR_ACCESS_DENIED = 0x5
ERROR_BAD_LENGTH = 0x18
ERROR_INVALID_PARAMETER = 0x57
ERROR_ALREADY_EXISTS = 0xB7


def get_last_error():
    return ctypes.GetLastError()


class UNICODE_STRING(Structure):
    _fields_ = [('length', c_ushort),
                ('maximum_length', c_ushort),
                ('buffer', c_void_p)]


class GUID(Structure):
    _fields_ = [("Data1", c_ulong),
                ("Data2", c_ushort),
                ("Data3", c_ushort),
                ("Data4", c_ubyte * 8)]
    _GUID_REGEX = re.compile('{([0-9A-F]{8})-([0-9A-F]{4})-([0-9A-F]{4})-([0-9A-F]{2})([0-9A-F]{2})-'
                             '([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})'
                             '([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})}', re.I)

    def __init__(self, gs=None):
        if gs:
            match = self._GUID_REGEX.match(gs)
            g = [int(i, 16) for i in match.groups()]
            self.Data1 = g[0]
            self.Data2 = g[1]
            self.Data3 = g[2]
            for i in range(8):
                self.Data4[i] = g[3 + i]

    def __str__(self):
        return "{%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x}" % \
               (self.Data1, self.Data2, self.Data3,
                self.Data4[0], self.Data4[1],
                self.Data4[2], self.Data4[3], self.Data4[4],
                self.Data4[5], self.Data4[6], self.Data4[7])



# Copyright 2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cpython.ref cimport PyObject
from libc.stddef cimport wchar_t
from .windows cimport WCHAR, CHAR, BYTE, ULONGLONG, LONGLONG, ULONG, LONG, SHORT, USHORT, ntohs, htonl, inet_ntoa, \
    in_addr, FLOAT, DOUBLE, DWORD, INT32
from .string cimport wstring, sprintf
from cython.operator cimport dereference as deref

cdef extern from "python.h":
    PyObject* PyUnicode_FromString(const char* u) nogil
    PyObject* PyUnicode_FromWideChar (wchar_t* w, Py_ssize_t size) nogil
    wchar_t* PyUnicode_AsWideCharString(PyObject* unicode, Py_ssize_t* size) nogil
    long PyLong_AsLong(PyObject *obj) nogil
    PyObject* Py_BuildValue(char* format, ...) nogil

    PyObject* PyTuple_New(Py_ssize_t len) nogil
    PyObject* PyTuple_GetItem(PyObject* p, Py_ssize_t pos) nogil
    int PyTuple_SetItem(PyObject* p, Py_ssize_t pos, PyObject* o) nogil

    void Py_XDECREF(PyObject* o) nogil
    void Py_XINCREF(PyObject* o)

    void PyMem_Free(void *p) nogil


cdef inline PyObject* _unicode(wchar_t* wchars) nogil:
    return PyUnicode_FromWideChar(wchars, -1)


cdef inline PyObject* _ansi(char* chars) nogil:
    return PyUnicode_FromString(chars)


cdef inline PyObject* _unicodec(void* buf) nogil:
    return Py_BuildValue('u', (<WCHAR*>buf)[0])


cdef inline PyObject* _ansic(void* buf) nogil:
    return Py_BuildValue('s', (<CHAR*>buf)[0])


cdef inline PyObject* _i8(void* buf) nogil:
    return Py_BuildValue('h', (<CHAR*>buf)[0])


cdef inline PyObject* _u8(void* buf) nogil:
    return Py_BuildValue('b', (<BYTE*>buf)[0])


cdef inline PyObject* _u8_hex(void* buf) nogil:
    cdef char hx[200]
    sprintf(hx, "%02x", (<SHORT*>buf)[0])
    return _ansi(hx)


cdef inline PyObject* _i16_hex(void* buf) nogil:
    cdef char hx[200]
    sprintf(hx, "%02x", (<CHAR*>buf)[0])
    return _ansi(hx)


cdef inline PyObject* _i64_hex(void* buf) nogil:
    cdef char hx[200]
    sprintf(hx, "%02x", (<ULONGLONG*>buf)[0])
    return _ansi(hx)


cdef inline PyObject* _i64(void* buf) nogil:
    return Py_BuildValue('L', (<LONGLONG*>buf)[0])


cdef inline PyObject* _u64(void* buf) nogil:
    return Py_BuildValue('K', (<ULONGLONG*>buf)[0])


cdef inline PyObject* _i32(void* buf) nogil:
    return Py_BuildValue('i', (<INT32*>buf)[0])


cdef inline PyObject* _i32_hex(void* buf) nogil:
    cdef char hx[200]
    sprintf(hx, "0x%x", (<ULONG*>buf)[0])
    return _ansi(hx)


cdef inline PyObject* _u32(void* buf) nogil:
    return Py_BuildValue('k', (<DWORD*>buf)[0])


cdef inline PyObject* _i16(void* buf) nogil:
    return Py_BuildValue('h', (<SHORT*>buf)[0])


cdef inline PyObject* _u16(void* buf) nogil:
    return Py_BuildValue('h', (<USHORT*>buf)[0])


cdef inline PyObject* _float(void* buf) nogil:
    return Py_BuildValue('f', (<FLOAT*>buf)[0])


cdef inline PyObject* _double(void* buf) nogil:
    return Py_BuildValue('d', (<DOUBLE*>buf)[0])


cdef inline PyObject* _ntohs(void* buf) nogil:
    return Py_BuildValue('h', ntohs((<USHORT*>buf)[0]))


cdef inline PyObject* _wstring(wstring ws):
    return PyUnicode_FromWideChar(ws.data(), ws.size())


cdef inline wchar_t* _wchar_t(PyObject* o) nogil:
    cdef Py_ssize_t size
    return PyUnicode_AsWideCharString(o, &size)


cdef inline PyObject* ip_addr(void* buf) nogil:
    cdef in_addr addr
    addr.S_un.S_addr = (<ULONG *>buf)[0]
    return Py_BuildValue('s', inet_ntoa(addr))


cdef inline wstring deref_prop(prop_name):
    return deref(new wstring(_wchar_t(<PyObject*>prop_name)))


# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from enum import Enum

from fibratus.common import NA
from fibratus.kevent_types import SEND_SOCKET_TCPV4, SEND_SOCKET_UDPV4, RECV_SOCKET_TCPV4, RECV_SOCKET_UDPV4, \
    ACCEPT_SOCKET_TCPV4, CONNECT_SOCKET_TCPV4, DISCONNECT_SOCKET_TCPV4, RECONNECT_SOCKET_TCPV4

import fibratus.tcpip.ports as ports


def port_to_proto(port, l4_proto):
    if 'TCP' in l4_proto:
        return ports.IANA_PORTS_TCP[port] if port in ports.IANA_PORTS_TCP else NA
    else:
        return ports.IANA_PORTS_UDP[port] if port in ports.IANA_PORTS_UDP else NA


class IpVer(Enum):

    IPV4 = 0
    IPV6 = 1


class TcpIpParser(object):

    def __init__(self, kevent):
        """TCP/IP kernel event parser.

        Packages the TCP and UDP requests into a single
        kernel event.

        Parameters:
        ----------

        kevent: dict
            kernel event representing the UDP / TCP request
        """
        self._kevent = kevent

    def parse_tcpip(self, ketype, ktcpip):
        """Parses the TCP/IP kernel events.

        Parameters
        ----------

        ketype: tuple
            network kernel event
        ktcpip:
            kevent payload as forwarded from the collector

        """
        pid = ktcpip.pid
        ip_src = ktcpip.saddr
        ip_dst = ktcpip.daddr
        sport = ktcpip.sport
        dport = ktcpip.dport

        self._kevent.pid = pid

        if ketype in [SEND_SOCKET_TCPV4,
                      SEND_SOCKET_UDPV4,
                      RECV_SOCKET_TCPV4,
                      RECV_SOCKET_UDPV4]:
            # get the application layer protocol
            # associated with the tcp segment
            # or the udp datagram
            if ketype in [SEND_SOCKET_TCPV4,
                          RECV_SOCKET_TCPV4]:
                l4_proto = 'TCP'
            else:
                l4_proto = 'UDP'
            protocol = port_to_proto(dport, l4_proto)
            if protocol == NA:
                protocol = port_to_proto(sport, l4_proto)
            self._kevent.params = {'pid': pid,
                                   'ip_src': ip_src,
                                   'ip_dst': ip_dst,
                                   'sport': sport,
                                   'dport': dport,
                                   'packet_size': ktcpip.size,
                                   'l4_proto': l4_proto,
                                   'protocol': protocol}
        elif ketype == ACCEPT_SOCKET_TCPV4:
            self._kevent.params = dict(pid=pid, ip_src=ip_src, ip_dst=ip_dst,
                                       sport=sport,
                                       dport=dport,
                                       rwin=ktcpip.rcvwin,
                                       protocol=port_to_proto(sport, 'TCP'))

        elif ketype == CONNECT_SOCKET_TCPV4:
            self._kevent.params = dict(pid=pid,
                                       ip_src=ip_src,
                                       ip_dst=ip_dst,
                                       sport=sport,
                                       dport=dport,
                                       rwin=ktcpip.rcvwin,
                                       protocol=port_to_proto(dport, 'TCP'))
        elif ketype == DISCONNECT_SOCKET_TCPV4:
            self._kevent.params = dict(pid=pid,
                                       ip_src=ip_src,
                                       ip_dst=ip_dst,
                                       sport=sport,
                                       dport=dport)

        elif ketype == RECONNECT_SOCKET_TCPV4:
            self._kevent.params = dict(pid=pid,
                                       ip_src=ip_src,
                                       ip_dst=ip_dst,
                                       sport=sport,
                                       dport=dport)


# Copyright 2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from kstream.includes.python cimport PyTuple_SetItem, Py_BuildValue, PyTuple_New
from cpython.ref cimport PyObject
from kstream.includes.windows cimport UCHAR

cdef inline PyObject* build_ktuple(PyObject* kguid, UCHAR opcode) nogil:
    cdef PyObject* ktuple = PyTuple_New(2)
    cdef PyObject* oco = Py_BuildValue('b', opcode)
    PyTuple_SetItem(ktuple, 0, kguid)
    PyTuple_SetItem(ktuple, 1, oco)
    return ktuple



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class KTraceError(Exception):
    pass


class FibratusError(Exception):
    pass


class FilamentError(Exception):
    pass


class TermInitializationError(Exception):
    pass


class UnknownKeventTypeError(Exception):

    def __init__(self, kevent):
        Exception.__init__(self, '%s cannot be recognized as a valid kernel event'
                           % kevent)


class HandleEnumError(Exception):

    def __init__(self, status):
        Exception.__init__(self, 'Unable to enumerate handles. Error code %s'
                           % status)


class InvalidPayloadError(Exception):
    pass


class BindingError(Exception):
    pass




# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from ctypes import POINTER
from ctypes.wintypes import DWORD, BOOL, HANDLE, ULONG, PULONG, BYTE, PDWORD

from fibratus.apidefs.cdefs import *
from fibratus.apidefs.sys import malloc, free
import fibratus.apidefs.declarer as declarer


# process access rights
PROCESS_VM_READ = 0x0010
PROCESS_DUP_HANDLE = 0x0040
PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_QUERY_LIMITED_INFORMATION = 0x1000

# thread access rights
THREAD_QUERY_INFORMATION = 0x0040

# ZwQueryInformationProcess constants
PROCESS_BASIC_INFO = 0
PROCESS_IMAGE_FILENAME = 27


# PEB (Process Environment Block) structures
class LIST_ENTRY(Structure):
    pass
LIST_ENTRY._fields_ = [('flink', POINTER(LIST_ENTRY)), ('blink', POINTER(LIST_ENTRY))]


class PEB_LDR_DATA(Structure):
    _fields_ = [('reserved1', BYTE * 8),
                ('reserved2', BYTE * 3),
                ('in_memory_order_module_list', LIST_ENTRY)]


class RTL_USER_PROCESS_PARAMETERS(Structure):
    _fields_ = [('reserved1', BYTE * 16),
                ('reserved2', PVOID * 10),
                ('image_path_name', UNICODE_STRING),
                ('command_line', UNICODE_STRING)]


class PEB(Structure):
    _fields_ = [('reserved1', BYTE * 2),
                ('being_debugged', BYTE),
                ('reserved2', BYTE * 21),
                ('ldr', POINTER(PEB_LDR_DATA)),
                ('process_parameters', POINTER(RTL_USER_PROCESS_PARAMETERS)),
                ('reserved3', BYTE * 520),
                ('post_process_init_routine', PVOID),
                ('reserved4', BYTE * 136),
                ('session_id', ULONG)]


class PROCESS_BASIC_INFORMATION(Structure):
    _fields_ = [('reserved1', PVOID),
                ('peb_base_address', POINTER(PEB)),
                ('reserved2', PVOID * 2),
                ('unique_process_id', PULONG),
                ('inherited_from_unique_process_id', ULONG)]

open_process = declarer.declare(declarer.KERNEL, 'OpenProcess',
                                [DWORD, BOOL, DWORD],
                                HANDLE)

open_thread = declarer.declare(declarer.KERNEL, 'OpenThread',
                               [DWORD, BOOL, DWORD],
                               HANDLE)

_read_process_memory = declarer.declare(declarer.KERNEL, 'ReadProcessMemory',
                                        [HANDLE, LPVOID, LPVOID, SIZE_T, POINTER(SIZE_T)],
                                        BOOL)

zw_query_information_process = declarer.declare(declarer.NT, 'ZwQueryInformationProcess',
                                                [HANDLE, DWORD, PVOID, ULONG, PULONG],
                                                DWORD)
query_full_process_image_name = declarer.declare(declarer.KERNEL, 'QueryFullProcessImageNameW',
                                                 [HANDLE, DWORD, LPTSTR, PDWORD],
                                                 BOOL)

get_current_process = declarer.declare(declarer.KERNEL, 'GetCurrentProcess',
                                       [],
                                       HANDLE)
get_process_id_of_thread = declarer.declare(declarer.KERNEL, 'GetProcessIdOfThread',
                                            [HANDLE],
                                            DWORD)


def read_process_memory(process, chunk, size):
    """Reads a memory block from the process address space.
    """
    buff = malloc(size)
    status = _read_process_memory(process,
                                  chunk,
                                  buff,
                                  size,
                                  None)
    if status != ERROR_SUCCESS:
        return buff
    else:
        free(buff)



# Copyright 2015 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import atexit
import os
import sys
from datetime import datetime

from kstreamc import KEventStreamCollector
from pykwalify.errors import SchemaError

from fibratus.binding.yar import YaraBinding
from fibratus.errors import BindingError
from fibratus.image_meta import ImageMetaRegistry
from fibratus.output.aggregator import OutputAggregator
from fibratus.output.console import ConsoleOutput
from fibratus.output.elasticsearch import ElasticsearchOutput
from fibratus.output.fs import FsOutput
from fibratus.output.smtp import SmtpOutput
from logbook import Logger, FileHandler, StreamHandler

import fibratus.apidefs.etw as etw
from fibratus.common import DotD as ddict, panic
from fibratus.config import YamlConfig
from fibratus.context_switch import ContextSwitchRegistry
from fibratus.controller import KTraceController, KTraceProps
from fibratus.dll import DllRepository
from fibratus.fs import FsIO
from fibratus.handle import HandleRepository
from fibratus.kevent import KEvent
from fibratus.kevent_types import *
from fibratus.tcpip.tcpip import TcpIpParser
from fibratus.output.amqp import AmqpOutput
from fibratus.registry import HiveParser
from fibratus.thread import ThreadRegistry


class Fibratus(object):

    """Fibratus entrypoint.

    Setup the core components including the kernel
    event stream collector and the tracing controller.
    At this point the system handles are also being
    enumerated.

    """
    def __init__(self, filament, **kwargs):

        self._start = datetime.now()
        try:
            log_path = os.path.join(os.path.expanduser('~'), '.fibratus', 'fibratus.log')
            FileHandler(log_path, mode='w+').push_application()
            StreamHandler(sys.stdout, bubble=True).push_application()
        except PermissionError:
            panic("ERROR - Unable to open log file for writing due to permission error")

        self.logger = Logger(Fibratus.__name__)
        self.logger.info('Starting Fibratus...')

        self._config = YamlConfig()
        self.logger.info('Loading configuration from [%s]' % self._config.config_path)
        try:
            self._config.load()
        except SchemaError as e:
            panic('Invalid configuration file. %s' % e.msg)

        enable_cswitch = kwargs.pop('cswitch', False)

        self.kcontroller = KTraceController()
        self.ktrace_props = KTraceProps()
        self.ktrace_props.enable_kflags(cswitch=enable_cswitch)
        self.ktrace_props.logger_name = etw.KERNEL_LOGGER_NAME

        enum_handles = kwargs.pop('enum_handles', True)

        self.handle_repository = HandleRepository()
        self._handles = []
        # query for handles on the
        # start of the kernel trace
        if enum_handles:
            self.logger.info('Enumerating system handles...')
            self._handles = self.handle_repository.query_handles()
            self.logger.info('%s handles found' % len(self._handles))
            self.handle_repository.free_buffers()

        image_meta_config = self._config.image_meta
        self.image_meta_registry = ImageMetaRegistry(image_meta_config.enabled, image_meta_config.imports,
                                                     image_meta_config.file_info)

        self.thread_registry = ThreadRegistry(self.handle_repository, self._handles,
                                              self.image_meta_registry)

        self.kevt_streamc = KEventStreamCollector(etw.KERNEL_LOGGER_NAME.encode())
        skips = self._config.skips
        image_skips = skips.images if 'images' in skips else []
        if len(image_skips) > 0:
            self.logger.info("Adding skips for images %s" % image_skips)
            for skip in image_skips:
                self.kevt_streamc.add_skip(skip)

        self.kevent = KEvent(self.thread_registry)

        self._output_classes = dict(console=ConsoleOutput,
                                    amqp=AmqpOutput,
                                    smtp=SmtpOutput,
                                    elasticsearch=ElasticsearchOutput,
                                    fs=FsOutput)
        self._outputs = self._construct_outputs()
        self.output_aggregator = OutputAggregator(self._outputs)

        self._binding_classes = dict(yara=YaraBinding)
        self._bindings = self._construct_bindings()

        if filament:
            filament.logger = self.logger
            filament.do_output_accessors(self._outputs)
        self._filament = filament

        self.fsio = FsIO(self.kevent, self._handles)
        self.hive_parser = HiveParser(self.kevent, self.thread_registry)
        self.tcpip_parser = TcpIpParser(self.kevent)
        self.dll_repository = DllRepository(self.kevent)
        self.context_switch_registry = ContextSwitchRegistry(self.thread_registry, self.kevent)

        self.output_kevents = {}
        self.filters_count = 0

    def run(self):

        @atexit.register
        def _exit():
            self.stop_ktrace()

        self.kcontroller.start_ktrace(etw.KERNEL_LOGGER_NAME, self.ktrace_props)

        def on_kstream_open():
            if self._filament is None:
                delta = datetime.now() - self._start
                self.logger.info('Started in %sm:%02ds.%s' % (int(delta.total_seconds() / 60), delta.seconds,
                                                              int(delta.total_seconds() * 1000)))
            else:
                self.logger.info('Running [%s] filament...' % self._filament.name)
        self.kevt_streamc.set_kstream_open_callback(on_kstream_open)
        self._open_kstream()

    def _open_kstream(self):
        try:
            self.kevt_streamc.open_kstream(self._on_next_kevent)
        except Exception as e:
            self.logger.error(e)
        except KeyboardInterrupt:
            self.stop_ktrace()

    def _construct_outputs(self):
        """Instantiates output classes.

        Builds the dictionary with instances
        of the output classes.
        """
        outputs = {}
        output_configs = self._config.outputs
        if not output_configs:
            return outputs
        for output in output_configs:
            name = next(iter(list(output.keys())), None)
            if name and \
                    name in self._output_classes.keys():
                # get the output configuration
                # and instantiate its class
                output_config = output[name]
                self.logger.info("Deploying [%s] output - [%s]"
                                 % (name, {k: v for k, v in output_config.items()
                                           if 'password' not in k}))
                output_class = self._output_classes[name]
                outputs[name] = output_class(**output_config)
        return outputs

    def _construct_bindings(self):
        """Builds binding classes.

        :return: dict: dictionary with instances of the binding classes
        """
        bindings = {}
        binding_configs = self._config.bindings
        if not binding_configs:
            return bindings
        for b in binding_configs:
            name = next(iter(list(b.keys())), None)
            if name and \
                    name in self._binding_classes.keys():
                binding_config = b[name]
                self.logger.info("Starting [%s] binding - [%s]" %
                                 (name, binding_config))
                binding_class = self._binding_classes[name]
                try:
                    binding = binding_class(self._outputs, self.logger,
                                            **binding_config)
                    bindings[name] = binding
                except BindingError as e:
                    self.logger.error("Couldn't start [%s] binding. Reason: %s" %
                                      (name, e))
        return bindings

    def __find_binding(self, name):
        return self._bindings[name] if name in self._bindings else None

    def stop_ktrace(self):
        self.logger.info('Stopping fibratus...')
        if self._filament:
            self._filament.close()
        self.kcontroller.stop_ktrace(self.ktrace_props)
        self.kevt_streamc.close_kstream()

    def add_filters(self, kevent_filters, **kwargs):
        self.kevt_streamc.add_pid_filter(kwargs.pop('pid', None))
        self.kevt_streamc.add_image_filter(kwargs.pop('image', None))
        if len(kevent_filters) > 0:
            self.filters_count = len(kevent_filters)
            # include the basic filters
            # that are essential to the
            # rest of kernel events
            self.kevt_streamc.add_ktuple_filter(ENUM_PROCESS)
            self.kevt_streamc.add_ktuple_filter(ENUM_THREAD)
            self.kevt_streamc.add_ktuple_filter(ENUM_IMAGE)
            self.kevt_streamc.add_ktuple_filter(REG_CREATE_KCB)
            self.kevt_streamc.add_ktuple_filter(REG_DELETE_KCB)

            # these kevents are necessary for consistent state
            # of the trace. If the user doesn't include them
            # in a filter list, then we do the job but set the
            # kernel event type as not eligible for rendering
            if KEvents.CREATE_PROCESS not in kevent_filters:
                self.kevt_streamc.add_ktuple_filter(CREATE_PROCESS)
                self.output_kevents[CREATE_PROCESS] = False
            else:
                self.output_kevents[CREATE_PROCESS] = True

            if KEvents.CREATE_THREAD not in kevent_filters:
                self.kevt_streamc.add_ktuple_filter(CREATE_THREAD)
                self.output_kevents[CREATE_THREAD] = False
            else:
                self.output_kevents[CREATE_THREAD] = True

            if KEvents.TERMINATE_PROCESS not in kevent_filters:
                self.kevt_streamc.add_ktuple_filter(TERMINATE_PROCESS)
                self.output_kevents[TERMINATE_PROCESS] = False
            else:
                self.output_kevents[TERMINATE_PROCESS] = True

            if KEvents.TERMINATE_THREAD not in kevent_filters:
                self.kevt_streamc.add_ktuple_filter(TERMINATE_THREAD)
                self.output_kevents[TERMINATE_THREAD] = False
            else:
                self.output_kevents[TERMINATE_THREAD] = True

            for kevent_filter in kevent_filters:
                ktuple = kname_to_tuple(kevent_filter)
                if isinstance(ktuple, list):
                    for kt in ktuple:
                        self.kevt_streamc.add_ktuple_filter(kt)
                        if kt not in self.output_kevents:
                            self.output_kevents[kt] = True
                else:
                    self.kevt_streamc.add_ktuple_filter(ktuple)
                    if ktuple not in self.output_kevents:
                        self.output_kevents[ktuple] = True

    def _on_next_kevent(self, ktype, cpuid, ts, kparams):
        """Callback which fires when new kernel event arrives.

        This callback is invoked for every new kernel event
        forwarded from the kernel stream collector.

        Parameters
        ----------

        ktype: tuple
            Kernel event type.
        cpuid: int
            Indentifies the CPU core where the event
            has been captured.
        ts: str
            Temporal reference of the kernel event.
        kparams: dict
            Kernel event's parameters.
        """

        # initialize kernel event properties
        self.kevent.ts = ts
        self.kevent.cpuid = cpuid
        self.kevent.name = ktuple_to_name(ktype)
        kparams = ddict(kparams)

        # thread / process kernel events
        if ktype in [CREATE_PROCESS,
                     CREATE_THREAD,
                     ENUM_PROCESS,
                     ENUM_THREAD]:
            self.thread_registry.add_thread(ktype, kparams)
            if ktype in [CREATE_PROCESS, CREATE_THREAD]:
                self.thread_registry.init_thread_kevent(self.kevent,
                                                        ktype,
                                                        kparams)
                # apply yara binding by matching against the process's image path
                if ktype == CREATE_PROCESS:
                    yara_binding = self.__find_binding('yara')
                    pid = int(kparams.process_id, 16)
                    thread = self.thread_registry.get_thread(pid)
                    if thread and yara_binding:
                        yara_binding.run(thread_info=thread,
                                         kevent=self.kevent)
                self._aggregate(ktype)

        elif ktype in [TERMINATE_PROCESS, TERMINATE_THREAD]:
            self.thread_registry.init_thread_kevent(self.kevent,
                                                    ktype,
                                                    kparams)
            self._aggregate(ktype)
            self.thread_registry.remove_thread(ktype, kparams)

        # file system/disk kernel events
        elif ktype in [CREATE_FILE,
                       DELETE_FILE,
                       CLOSE_FILE,
                       READ_FILE,
                       WRITE_FILE,
                       RENAME_FILE,
                       SET_FILE_INFORMATION]:
            self.fsio.parse_fsio(ktype, kparams)
            self._aggregate(ktype)

        # dll kernel events
        elif ktype in [LOAD_IMAGE, ENUM_IMAGE]:
            self.dll_repository.register_dll(kparams)
            if ktype == LOAD_IMAGE:
                self._aggregate(ktype)
        elif ktype == UNLOAD_IMAGE:
            self.dll_repository.unregister_dll(kparams)
            self._aggregate(ktype)
        #
        # # registry kernel events
        elif ktype == REG_CREATE_KCB:
            self.hive_parser.add_kcb(kparams)
        elif ktype == REG_DELETE_KCB:
            self.hive_parser.remove_kcb(kparams.key_handle)

        elif ktype in [REG_CREATE_KEY,
                       REG_DELETE_KEY,
                       REG_OPEN_KEY,
                       REG_QUERY_KEY,
                       REG_SET_VALUE,
                       REG_DELETE_VALUE,
                       REG_QUERY_VALUE]:
            self.hive_parser.parse_hive(ktype, kparams)
            self._aggregate(ktype)

        # network kernel events
        elif ktype in [SEND_SOCKET_TCPV4,
                       SEND_SOCKET_UDPV4,
                       RECV_SOCKET_TCPV4,
                       RECV_SOCKET_UDPV4,
                       ACCEPT_SOCKET_TCPV4,
                       CONNECT_SOCKET_TCPV4,
                       DISCONNECT_SOCKET_TCPV4,
                       RECONNECT_SOCKET_TCPV4]:
            self.tcpip_parser.parse_tcpip(ktype, kparams)
            self._aggregate(ktype)

        # context switch events
        elif ktype == CONTEXT_SWITCH:
            self.context_switch_registry.next_cswitch(cpuid, ts, kparams)
            self._aggregate(ktype)

        if self._filament:
            if ktype not in [ENUM_PROCESS,
                             ENUM_THREAD,
                             ENUM_IMAGE,
                             REG_CREATE_KCB,
                             REG_DELETE_KCB]:
                ok = self.output_kevents[ktype] if ktype in self.output_kevents \
                    else False
                if self.kevent.name and ok:
                    thread = self.kevent.thread
                    kevent = {
                        'params': self.kevent.params,
                        'name': self.kevent.name,
                        'pid': self.kevent.pid,
                        'tid': self.kevent.tid,
                        'timestamp': self.kevent.ts,
                        'cpuid': self.kevent.cpuid,
                        'category': self.kevent.category
                    }
                    if thread:
                        kevent.update({
                            'thread': {
                                'name': thread.name,
                                'exe': thread.exe,
                                'comm': thread.comm,
                                'pid': thread.pid,
                                'ppid': thread.ppid
                            }
                        })
                    self._filament.on_next_kevent(kevent)

    def _aggregate(self, ktype):
        """Aggregates the kernel event to the output sink.

        Parameters
        ----------

        ktype: tuple
            Identifier of the kernel event
        """
        if not self._filament:
            if ktype in self.output_kevents:
                if self.output_kevents[ktype]:
                    self.kevent.inc_kid()
                    self.output_aggregator.aggregate(self.kevent)
            elif self.filters_count == 0:
                self.kevent.inc_kid()
                self.output_aggregator.aggregate(self.kevent)


# Copyright 2015/2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest.mock import patch, call

import pytest

from fibratus.apidefs.process import PROCESS_DUP_HANDLE
from fibratus.handle import HandleRepository, HandleType
from fibratus.common import DotD as dd


def raw_handles():
    handles = {}
    handles[18446738026470543136] = dd(obj=18446738026470543136, handle=12, pid=3472, access_mask=1048608, obj_type_index=28)
    handles[18446738026474344592] = dd(obj=18446738026474344592, handle=16, pid=3472, access_mask=2031617, obj_type_index=36)
    handles[18446738026474569824] = dd(obj=18446738026474569824, handle=204, pid=920, access_mask=1048578, obj_type_index=17)
    handles[18446738026469227424] = dd(obj=18446738026469227424, handle=116, pid=920, access_mask=2031619, obj_type_index=12)
    handles[18446735964891181152] = dd(obj=18446735964891181152, handle=108, pid=1616, access_mask=983551, obj_type_index=5)
    return handles


def query_handle_side_effects():
    return [dd(contents=dd(type_name=dd(buffer="FILE"))), dd(contents=dd(buffer="\Device\HarddiskVolume2"))] \
        * len(raw_handles())


@pytest.fixture(scope='module')
def handle_repo():
    return HandleRepository()


class TestHandleRepository():

    def test_init_handle_repository(self, handle_repo):
        handle_types = handle_repo._handle_types
        assert isinstance(handle_types, list)
        assert HandleType.FILE.name in handle_types

    @patch('fibratus.handle.HandleRepository._enum_handles', return_value=raw_handles())
    @patch('fibratus.handle.open_process', side_effect=[None, 200, 301, 120, 343])
    @patch('fibratus.handle.duplicate_handle', side_effect=[1, 1, 1, 0])
    @patch('fibratus.handle.HandleRepository._query_handle', side_effect=query_handle_side_effects())
    @patch('fibratus.handle.close_handle')
    @patch('fibratus.handle.get_current_process', return_value=os.getpid())
    @patch('fibratus.handle.cast', return_value=dd(value='FILE'))
    @patch('fibratus.handle.byref', return_value=0x100)
    def test_query_handles(self, byref_mock, cast_mock, get_current_process_mock, close_handle_mock,
                           query_handle_mock, duplicate_handle_mock, open_process_mock,
                           enum_handles_mock, handle_repo):
        handles = handle_repo.query_handles()
        assert enum_handles_mock.called
        assert get_current_process_mock.called

        open_process_expected_calls = [call(PROCESS_DUP_HANDLE, False, 3472), call(PROCESS_DUP_HANDLE, False, 3472),
                                       call(PROCESS_DUP_HANDLE, False, 920),
                                       call(PROCESS_DUP_HANDLE, False, 920),
                                       call(PROCESS_DUP_HANDLE, False, 1616)]
        open_process_mock.assert_has_calls(open_process_expected_calls, any_order=True)

        assert duplicate_handle_mock.call_count == 4
        assert query_handle_mock.call_count == 6
        assert close_handle_mock.call_count == 7

        assert len(handles) == 3
        assert len([h for h in handles if h.handle_type == HandleType.FILE]) > 0


# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime
from enum import Enum

from fibratus.apidefs.process import open_thread, THREAD_QUERY_INFORMATION, get_process_id_of_thread
from fibratus.apidefs.sys import close_handle
from fibratus.common import DotD as ddict, NA


class Category(Enum):

    REGISTRY = 0
    FILE = 1
    NET = 2
    PROCESS = 3
    THREAD = 4
    MM = 5
    CSWITCH = 6
    SYSCALL = 7
    DISK_IO = 8
    DLL = 9
    OTHER = 10


class KEvents(object):
    """Available kernel event names.
    """
    CREATE_PROCESS = 'CreateProcess'
    CREATE_THREAD = 'CreateThread'
    TERMINATE_PROCESS = 'TerminateProcess'
    TERMINATE_THREAD = 'TerminateThread'

    REG_CREATE_KEY = 'RegCreateKey'
    REG_DELETE_KEY = 'RegDeleteKey'
    REG_DELETE_VALUE = 'RegDeleteValue'
    REG_OPEN_KEY = 'RegOpenKey'
    REG_SET_VALUE = 'RegSetValue'
    REG_QUERY_VALUE = 'RegQueryValue'
    REG_QUERY_KEY = 'RegQueryKey'

    CREATE_FILE = 'CreateFile'
    DELETE_FILE = 'DeleteFile'
    WRITE_FILE = 'WriteFile'
    READ_FILE = 'ReadFile'
    CLOSE_FILE = 'CloseFile'
    RENAME_FILE = 'RenameFile'
    SET_FILE_INFORMATION = 'SetFileInformation'

    SEND = 'Send'
    RECEIVE = 'Recv'
    ACCEPT = 'Accept'
    CONNECT = 'Connect'
    DISCONNECT = 'Disconnect'
    RECONNECT = 'Reconnect'

    LOAD_IMAGE = 'LoadImage'
    UNLOAD_IMAGE = 'UnloadImage'

    SYSCALL_ENTER = 'SyscallEnter'
    SYSCALL_EXIT = 'SyscallExit'

    CONTEXT_SWITCH = 'ContextSwitch'

    @classmethod
    def all(cls):
        return [cls.CREATE_PROCESS,
                cls.CREATE_THREAD,
                cls.TERMINATE_PROCESS,
                cls.TERMINATE_THREAD,
                cls.CREATE_FILE,
                cls.DELETE_FILE,
                cls.READ_FILE,
                cls.WRITE_FILE,
                cls.CLOSE_FILE,
                cls.RENAME_FILE,
                cls.SET_FILE_INFORMATION,
                cls.REG_QUERY_KEY,
                cls.REG_QUERY_VALUE,
                cls.REG_CREATE_KEY,
                cls.REG_DELETE_KEY,
                cls.REG_DELETE_VALUE,
                cls.REG_OPEN_KEY,
                cls.REG_SET_VALUE,
                cls.LOAD_IMAGE,
                cls.UNLOAD_IMAGE,
                cls.SEND,
                cls.RECEIVE,
                cls.ACCEPT,
                cls.CONNECT,
                cls.RECONNECT,
                cls.DISCONNECT,
                cls.CONTEXT_SWITCH]

    @classmethod
    def meta_info(cls):
        kevents = {
            KEvents.CREATE_PROCESS: (Category.PROCESS, 'Creates a new process and its primary thread', ),
            KEvents.CREATE_THREAD: (Category.THREAD, 'Creates a thread to execute within the virtual address space'
                                                     ' of the calling process', ),
            KEvents.TERMINATE_PROCESS: (Category.PROCESS, 'Terminates the process and all of its threads', ),
            KEvents.TERMINATE_THREAD: (Category.THREAD, 'Terminates a thread', ),
            KEvents.CREATE_FILE: (Category.FILE, 'Creates or opens a file or I/O device', ),
            KEvents.DELETE_FILE: (Category.FILE, 'Deletes an existing file or directory', ),
            KEvents.READ_FILE: (Category.FILE, 'Reads data from the file or I/O device', ),
            KEvents.WRITE_FILE: (Category.FILE, 'Writes data to the file or I/O device', ),
            KEvents.CLOSE_FILE: (Category.FILE, 'Closes the file or I/O device', ),
            KEvents.SET_FILE_INFORMATION: (Category.FILE, 'Changes information for the specified file',),
            KEvents.RENAME_FILE: (Category.FILE, 'Renames a file or directory', ),
            KEvents.REG_QUERY_KEY: (Category.REGISTRY, 'Retrieves information about the registry key', ),
            KEvents.REG_OPEN_KEY: (Category.REGISTRY, 'Opens the registry key', ),
            KEvents.REG_CREATE_KEY: (Category.REGISTRY, 'Creates the registry key or open it if the key '
                                                        'already exists', ),
            KEvents.REG_DELETE_KEY: (Category.REGISTRY, 'Deletes a subkey and its values', ),
            KEvents.REG_QUERY_VALUE: (Category.REGISTRY, 'Retrieves the type and data of the value'
                                                         ' associated with an open registry key', ),
            KEvents.REG_DELETE_VALUE: (Category.REGISTRY, 'Removes a value from the registry key', ),
            KEvents.REG_SET_VALUE: (Category.REGISTRY, 'Sets the data and type of a value under a registry key', ),
            KEvents.LOAD_IMAGE: (Category.DLL, 'Loads the module into the address space of the calling process', ),
            KEvents.UNLOAD_IMAGE: (Category.DLL, 'Frees the loaded module from the address space '
                                                 'of the calling process', ),
            KEvents.SEND: (Category.NET, 'Sends data on a connected socket', ),
            KEvents.RECEIVE: (Category.NET, 'Receives data from a connected socket', ),
            KEvents.ACCEPT: (Category.NET, 'Initiates the connection attempt from the remote or local TCP socket', ),
            KEvents.CONNECT: (Category.NET, 'Establishes the connection to a TCP socket', ),
            KEvents.RECONNECT: (Category.NET, 'Reconnects to a TCP socket', ),
            KEvents.DISCONNECT: (Category.NET, 'Closes the connection to a TCP socket', ),

            KEvents.CONTEXT_SWITCH: (Category.THREAD, 'Scheduler selects a new thread to execute',)}
        return kevents

__kevents__ = KEvents.meta_info()


class KEvent(object):

    def __init__(self, thread_registry):
        self._kid = 0
        self._ts = datetime.now()
        self._cpuid = 0
        self._name = None
        self._category = None
        self._params = {}
        self._tid = None
        self._pid = None
        self.thread_registry = thread_registry

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name
        if name in __kevents__:
            cat, _ = __kevents__[name]
            self._category = cat.name

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, params):
        self._params = ddict(params)

    @property
    def ts(self):
        return self._ts

    @ts.setter
    def ts(self, ts):
        self._ts = datetime.strptime(ts, '%Y-%m-%d %H:%M:%S.%f')

    @property
    def cpuid(self):
        return self._cpuid

    @cpuid.setter
    def cpuid(self, cpuid):
        self._cpuid = cpuid

    @property
    def category(self):
        return self._category

    @property
    def pid(self):
        return self._pid

    @pid.setter
    def pid(self, pid):
        self._pid = pid

    @property
    def tid(self):
        return self._tid

    @tid.setter
    def tid(self, tid):
        self._tid = tid

    @property
    def kid(self):
        return self._kid

    @property
    def thread(self):
        return self._find_thread()

    def _find_thread(self):
        """Finds the current thread/process emitted by the kernel event.
        """
        if self._pid:
            # first lookup by process id
            # if the process doesn't exist
            # in the thread registry
            # then query by the thread id
            thread = self.thread_registry.get_thread(self._pid)
            if not thread and self._tid:
                thread = self.thread_registry.get_thread(self._tid)
        else:
            # we dont have the process id
            # try to find the thread from which
            # we can get the process
            thread = self.thread_registry.get_thread(self._tid)
        return thread

    def get_thread(self):
        """Gets the thread associated with the kernel event.
        """
        thread = self._find_thread()
        if thread:
            return thread.pid, thread.name
        else:
            # figure out the process id from thread
            # if the process can't be found in
            # the thread registry
            pid = NA
            if self._pid is None:
                if self._tid:
                    # get the thread handle
                    handle = open_thread(THREAD_QUERY_INFORMATION,
                                         False,
                                         self._tid)
                    if handle:
                        pid = get_process_id_of_thread(handle)
                        close_handle(handle)
            else:
                pid = self._pid
            return pid, NA

    def inc_kid(self):
        self._kid += 1





# Copyright 2015 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from fibratus.errors import UnknownKeventTypeError
from fibratus.kevent import KEvents


# start process event
CREATE_PROCESS = ('{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}', 1)
# end process event
TERMINATE_PROCESS = ('{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}', 2)
# enum processes event
ENUM_PROCESS = ('{3d6fa8d0-fe05-11d0-9dda-00c04fd7ba7c}', 3)
# start thread event
CREATE_THREAD = ('{3d6fa8d1-fe05-11d0-9dda-00c04fd7ba7c}', 1)
# end thread event
TERMINATE_THREAD = ('{3d6fa8d1-fe05-11d0-9dda-00c04fd7ba7c}', 2)
# enum threads event
ENUM_THREAD = ('{3d6fa8d1-fe05-11d0-9dda-00c04fd7ba7c}', 3)

# create file event
CREATE_FILE = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 64)
# delete file event
DELETE_FILE = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 70)
# close file event generated when the file object is freed
CLOSE_FILE = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 66)
# read file event
READ_FILE = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 67)
# write file event
WRITE_FILE = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 68)
# rename file event
RENAME_FILE = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 71)
# enumerate directory event
ENUM_DIRECTORY = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 72)
# set file information event
SET_FILE_INFORMATION = ('{90cbdc39-4a3e-11d1-84f4-0000f80464e3}', 69)

# disk read event
DISK_IO_READ = ('{3d6fa8d4-fe05-11d0-9dda-00c04fd7ba7c}', 10)
# disk write event
DISK_IO_WRITE = ('{3d6fa8d4-fe05-11d0-9dda-00c04fd7ba7c}', 11)

# create registry key event
REG_CREATE_KEY = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 10)
# create registry key event
REG_DELETE_KEY = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 12)
# delete registry value event
REG_DELETE_VALUE = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 15)
# registry open key
REG_OPEN_KEY = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 11)
# registry set value key event
REG_SET_VALUE = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 14)
# registry query value key event
REG_QUERY_VALUE = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 16)
# registry query value key event
REG_QUERY_KEY = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 13)
# create the key control block
REG_CREATE_KCB = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 22)
# delete the key control block
REG_DELETE_KCB = ('{ae53722e-c863-11d2-8659-00c04fa321a1}', 23)

# image load event generated when a DLL or executable file is loaded
LOAD_IMAGE = ('{2cb15d1d-5fc1-11d2-abe1-00a0c911f518}', 10)
# generated when a DLL or executable file is unloaded
UNLOAD_IMAGE = ('{2cb15d1d-5fc1-11d2-abe1-00a0c911f518}', 2)
# enumerates all loaded images
ENUM_IMAGE = ('{2cb15d1d-5fc1-11d2-abe1-00a0c911f518}', 3)

# virtual memory allocation event
VIRTUAL_ALLOC = ('{3d6fa8d3-fe05-11d0-9dda-00c04fd7ba7c}', 98)
# virtual memory free event
VIRTUAL_FREE = ('{3d6fa8d3-fe05-11d0-9dda-00c04fd7ba7c}', 99)

# system call enter event
SYSCALL_ENTER = ('{ce1dbfb4-137e-4da6-87b0-3f59aa102cbc}', 51)
# system call exit event
SYSCALL_EXIT = ('{ce1dbfb4-137e-4da6-87b0-3f59aa102cbc}', 52)

# context switch event
CONTEXT_SWITCH = ('{3d6fa8d1-fe05-11d0-9dda-00c04fd7ba7c}', 36)

# starts an incoming connection attempt on socket
ACCEPT_SOCKET_TCPV4 = ('{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}', 15)
ACCEPT_SOCKET_TCPV6 = ('{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}', 31)

# sends data on a connected socket
SEND_SOCKET_TCPV4 = ('{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}', 10)
SEND_SOCKET_UDPV4 = ('{bf3a50c5-a9c9-4988-a005-2df0b7c80f80}', 10)
# establishes a connection to a specified socket
CONNECT_SOCKET_TCPV4 = ('{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}', 12)
# disconnect event
DISCONNECT_SOCKET_TCPV4 = ('{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}', 13)
# reconnect attempt event
RECONNECT_SOCKET_TCPV4 = ('{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}', 16)
# receives data from a connected socket
RECV_SOCKET_TCPV4 = ('{9a280ac0-c8e0-11d1-84e2-00c04fb998a2}', 11)
RECV_SOCKET_UDPV4 = ('{bf3a50c5-a9c9-4988-a005-2df0b7c80f80}', 11)


def kname_to_tuple(name):

    if name == KEvents.CREATE_PROCESS:
        return CREATE_PROCESS
    elif name == KEvents.TERMINATE_PROCESS:
        return TERMINATE_PROCESS
    elif name == KEvents.CREATE_THREAD:
        return CREATE_THREAD
    elif name == KEvents.TERMINATE_THREAD:
        return TERMINATE_THREAD

    elif name == KEvents.REG_CREATE_KEY:
        return REG_CREATE_KEY
    elif name == KEvents.REG_QUERY_KEY:
        return REG_QUERY_KEY
    elif name == KEvents.REG_OPEN_KEY:
        return REG_OPEN_KEY
    elif name == KEvents.REG_QUERY_VALUE:
        return REG_QUERY_VALUE
    elif name == KEvents.REG_SET_VALUE:
        return REG_SET_VALUE
    elif name == KEvents.REG_DELETE_KEY:
        return REG_DELETE_KEY
    elif name == KEvents.REG_DELETE_VALUE:
        return REG_DELETE_VALUE

    elif name == KEvents.CREATE_FILE:
        return CREATE_FILE
    elif name == KEvents.READ_FILE:
        return READ_FILE
    elif name == KEvents.WRITE_FILE:
        return WRITE_FILE
    elif name == KEvents.CLOSE_FILE:
        return CLOSE_FILE
    elif name == KEvents.DELETE_FILE:
        return DELETE_FILE
    elif name == KEvents.RENAME_FILE:
        return RENAME_FILE
    elif name == KEvents.SET_FILE_INFORMATION:
        return SET_FILE_INFORMATION

    elif name == KEvents.LOAD_IMAGE:
        return LOAD_IMAGE
    elif name == KEvents.UNLOAD_IMAGE:
        return UNLOAD_IMAGE

    elif name == KEvents.CONTEXT_SWITCH:
        return CONTEXT_SWITCH

    elif name == KEvents.SEND:
        return [SEND_SOCKET_UDPV4, SEND_SOCKET_TCPV4]
    elif name == KEvents.RECEIVE:
        return [RECV_SOCKET_UDPV4, RECV_SOCKET_TCPV4]
    elif name == KEvents.ACCEPT:
        return [ACCEPT_SOCKET_TCPV4, ACCEPT_SOCKET_TCPV6]
    elif name == KEvents.CONNECT:
        return CONNECT_SOCKET_TCPV4
    elif name == KEvents.RECONNECT:
        return RECONNECT_SOCKET_TCPV4
    elif name == KEvents.DISCONNECT:
        return DISCONNECT_SOCKET_TCPV4
    else:
        raise UnknownKeventTypeError(name)


def ktuple_to_name(ktuple):

    if ktuple == CREATE_PROCESS:
        return KEvents.CREATE_PROCESS
    elif ktuple == CREATE_THREAD:
        return KEvents.CREATE_THREAD
    elif ktuple == TERMINATE_PROCESS:
        return KEvents.TERMINATE_PROCESS
    elif ktuple == TERMINATE_THREAD:
        return KEvents.TERMINATE_THREAD

    elif ktuple == REG_CREATE_KEY:
        return KEvents.REG_CREATE_KEY
    elif ktuple == REG_DELETE_KEY:
        return KEvents.REG_DELETE_KEY
    elif ktuple == REG_DELETE_VALUE:
        return KEvents.REG_DELETE_VALUE
    elif ktuple == REG_OPEN_KEY:
        return KEvents.REG_OPEN_KEY
    elif ktuple == REG_SET_VALUE:
        return KEvents.REG_SET_VALUE
    elif ktuple == REG_QUERY_VALUE:
        return KEvents.REG_QUERY_VALUE
    elif ktuple == REG_QUERY_KEY:
        return KEvents.REG_QUERY_KEY

    elif ktuple == CREATE_FILE:
        return KEvents.CREATE_FILE
    elif ktuple == DELETE_FILE:
        return KEvents.DELETE_FILE
    elif ktuple == CLOSE_FILE:
        return KEvents.CLOSE_FILE
    elif ktuple == WRITE_FILE:
        return KEvents.WRITE_FILE
    elif ktuple == READ_FILE:
        return KEvents.READ_FILE
    elif ktuple == RENAME_FILE:
        return KEvents.RENAME_FILE
    elif ktuple == SET_FILE_INFORMATION:
        return KEvents.SET_FILE_INFORMATION

    elif ktuple == LOAD_IMAGE:
        return KEvents.LOAD_IMAGE
    elif ktuple == UNLOAD_IMAGE:
        return KEvents.UNLOAD_IMAGE

    elif ktuple == CONTEXT_SWITCH:
        return KEvents.CONTEXT_SWITCH

    elif ktuple == SEND_SOCKET_UDPV4 or\
            ktuple == SEND_SOCKET_TCPV4:
        return KEvents.SEND
    elif ktuple == RECV_SOCKET_UDPV4 or\
            ktuple == RECV_SOCKET_TCPV4:
        return KEvents.RECEIVE
    elif ktuple == ACCEPT_SOCKET_TCPV4:
        return KEvents.ACCEPT
    elif ktuple == CONNECT_SOCKET_TCPV4:
        return KEvents.CONNECT
    elif ktuple == DISCONNECT_SOCKET_TCPV4:
        return KEvents.DISCONNECT
    elif ktuple == RECONNECT_SOCKET_TCPV4:
        return KEvents.RECONNECT


# Copyright 2015/2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Triggers when a process creates the registry value which
would enable it to execute on system startup.
"""


keys = ['Run', 'RunOnce', 'RunServices', 'RunServicesOnce', 'Userinit']


def on_init():
    set_filter('RegSetValue')


def on_next_kevent(kevent):
    if kevent.thread:
        process_name = kevent.thread.name
        key = kevent.params.key
        if key in keys:
            message = 'The process %s has created a ' \
                      'persistent registry value , ' \
                      'under %s with content %s' \
                       % (process_name,
                          '%s/%s' % (kevent.params.hive, key),
                          kevent.params.value)
            smtp.emit(message, subject='Registry persistence detected')



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Usage:
    fibratus run ([--filament=<filament>] | [--filters <kevents>...])
                 [--pid=<pid> | --image=<image>] [--no-enum-handles] [--cswitch]
    fibratus list-kevents
    fibratus list-filaments
    fibratus -h | --help
    fibratus --version

Options:
    -h --help                 Show this screen.
    --filament=<filament>     Specify the filament to execute.
    --no-enum-handles         Avoids enumerating the system handles.
    --pid=<pid>               Spy on a specific process identifier.
    --image=<image>           Spy on a specific image name.
    --cswitch                 Enables context switch kernel events.
    --version                 Show version.
"""
import sys

from docopt import docopt

from fibratus.apidefs.sys import set_console_ctrl_handler, PHANDLER_ROUTINE
from fibratus.errors import FilamentError
from fibratus.entrypoint import Fibratus
from fibratus.filament import Filament
from fibratus.kevent import KEvents
from fibratus.version import VERSION
from fibratus.common import panic, Tabular

args = docopt(__doc__, version=VERSION)

kevent_filters = args['<kevents>']
filament_name = args['--filament'] if args['--filament'] else None


def _check_kevent(kevent):
    if kevent not in KEvents.all():
        panic('fibratus run: ERROR - %s is not a valid kernel event. Run list-kevents to see '
              'the available kernel events' % kevent)


def main():
    if args['run']:
        if len(kevent_filters) > 0 and not filament_name:
            for kfilter in kevent_filters:
                _check_kevent(kfilter)

        enum_handles = False if args['--no-enum-handles'] else True
        cswitch = True if args['--cswitch'] else False

        filament = None
        filament_filters = []

        if filament_name:
            if not Filament.exists(filament_name):
                panic('fibratus run: ERROR - %s filament does not exist. Run list-filaments to see '
                      'the available filaments' % filament_name)
            filament = Filament()
            try:
                filament.load_filament(filament_name)
            except FilamentError as e:
                panic('fibratus run: ERROR - %s' % e)

            filament_filters = filament.filters

            if len(filament_filters) > 0:
                for kfilter in filament_filters:
                    _check_kevent(kfilter)

            filament.render_tabular()

        try:
            fibratus = Fibratus(filament, enum_handles=enum_handles, cswitch=cswitch)
        except KeyboardInterrupt:
            # the user has stopped command execution
            # before opening the kernel event stream
            sys.exit(0)

        @PHANDLER_ROUTINE
        def handle_ctrl_c(event):
            if event == 0:
                fibratus.stop_ktrace()
            return 0
        set_console_ctrl_handler(handle_ctrl_c, True)

        # add specific filters
        filters = dict()
        filters['pid'] = args['--pid'] if args['--pid'] else None
        filters['image'] = args['--image'] if args['--image'] else None

        if not filament:
            if len(kevent_filters) > 0:
                fibratus.add_filters(kevent_filters, **filters)
            else:
                fibratus.add_filters([], **filters)
        else:
            if len(filament_filters) > 0:
                fibratus.add_filters(filament_filters, **filters)
            else:
                fibratus.add_filters([], **filters)
        try:
            fibratus.run()
        except KeyboardInterrupt:
            set_console_ctrl_handler(handle_ctrl_c, False)

    elif args['list-filaments']:
        filaments = Tabular(['Filament', 'Description'], 'Description',
                            sort_by='Filament')
        for filament, desc in Filament.list_filaments().items():
            filaments.add_row([filament, desc])
        filaments.draw()

    elif args['list-kevents']:
        kevents = Tabular(['KEvent', 'Category', 'Description'], 'Description',
                          sort_by='Category')
        for kevent, meta in KEvents.meta_info().items():
            kevents.add_row([kevent, meta[0].name, meta[1]])
        kevents.draw()



# Copyright 2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import Mock, call

from datetime import datetime
import pytest

from fibratus.context_switch import ContextSwitchRegistry, ThreadState, WaitMode, WaitReason
from fibratus.kevent import KEvent
from fibratus.thread import ThreadRegistry, ThreadInfo
from fibratus.common import DotD as ddict


@pytest.fixture(scope='module')
def thread_registry_mock():
    thread_registry = Mock(spec_set=ThreadRegistry)
    thread_info1 = ThreadInfo(456, int('0x1054', 16), 22, 'explorer.exe', 'C:\\Windows\\EXPLORER.exe')
    thread_info2 = ThreadInfo(1, int('0x0', 16), 2, 'system.exe', 'C:\\Windows\\system.exe')
    thread_registry.get_thread.side_effect = [thread_info1, thread_info2]
    return thread_registry


@pytest.fixture(scope='module')
def thread_registry_empty_mock():
    return Mock(spec_set=ThreadRegistry)


@pytest.fixture(scope='module')
def kevent_mock():
    return Mock(spec_set=KEvent)


class TestContextSwitchRegistry(object):

    def test_next_cswitch(self, thread_registry_mock, kevent_mock):

        context_switch_registry = ContextSwitchRegistry(thread_registry_mock, kevent_mock)
        ts = datetime.strptime("12:05:45.233", '%H:%M:%S.%f')
        kcs1 = ddict({'old_thread_wait_ideal_processor': 2, 'previous_c_state': 1, 'old_thread_state': 2,
                      'old_thread_priority': 0, 'reserved': 777748717, 'spare_byte': 0,
                      'old_thread_wait_reason': 0, 'new_thread_wait_time': '0x0', 'old_thread_wait_mode': 0,
                      'new_thread_priority': 15, 'new_thread_id': '0x1054', 'old_thread_id': '0x0'})
        new_thread_id = int(kcs1.new_thread_id, 16)
        context_switch_registry.next_cswitch(1, ts, kcs1)

        thread_registry_mock.get_thread.assert_has_calls([call(new_thread_id),
                                                         call(int(kcs1.old_thread_id, 16))])

        assert (1, new_thread_id,) in context_switch_registry.context_switches()
        cs = context_switch_registry.context_switches()[(1, new_thread_id,)]
        assert cs

        assert cs.timestamp is ts
        assert cs.next_proc_name == "explorer.exe"
        assert cs.next_thread_wait_time == 0
        assert cs.next_thread_prio == 15
        assert cs.prev_thread_prio == 0
        assert cs.prev_thread_state is ThreadState.RUNNING
        assert cs.count == 1
        assert cs.prev_thread_wait_mode is WaitMode.KERNEL
        assert cs.prev_thread_wait_reason is WaitReason.EXECUTIVE

    def test_next_cswitch_in_registry(self, thread_registry_empty_mock, kevent_mock):
        context_switch_registry = ContextSwitchRegistry(thread_registry_empty_mock, kevent_mock)

        kcs1 = ddict({'old_thread_wait_ideal_processor': 3, 'previous_c_state': 0, 'old_thread_state': 5,
                      'old_thread_priority': 8, 'reserved': 4294967294, 'spare_byte': 0,
                      'old_thread_wait_reason': 8, 'new_thread_wait_time': '0x0', 'old_thread_wait_mode': 1,
                      'new_thread_priority': 8, 'new_thread_id': '0x1fc8', 'old_thread_id': '0x2348'})
        kcs2 = ddict({'old_thread_wait_ideal_processor': 3, 'previous_c_state': 0, 'old_thread_state': 3,
                      'old_thread_priority': 8, 'reserved': 4294967295, 'spare_byte': 0,
                      'old_thread_wait_reason': 8, 'new_thread_wait_time': '0x0', 'old_thread_wait_mode': 1,
                      'new_thread_priority': 5, 'new_thread_id': '0x1fc8', 'old_thread_id': '0x2348'})

        context_switch_registry.next_cswitch(1, datetime.strptime("12:05:45.233", '%H:%M:%S.%f'), kcs1)
        context_switch_registry.next_cswitch(1, datetime.strptime("12:05:45.234", '%H:%M:%S.%f'), kcs2)

        k = (1, int(kcs1.new_thread_id, 16))
        cs = context_switch_registry.context_switches()[k]
        assert cs
        assert cs.count == 2
        assert cs.next_thread_prio == 5
        assert cs.prev_thread_state is ThreadState.STANDBY
        assert cs.timestamp == datetime.strptime("12:05:45.234", '%H:%M:%S.%f')
        assert cs.prev_thread_wait_reason is WaitReason.FREE_PAGE



# Copyright 2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from libc.stddef cimport wchar_t

cdef extern from "windows.h":
    ctypedef unsigned long ULONG
    ctypedef unsigned char BYTE
    ctypedef unsigned long DWORD
    ctypedef signed int INT32
    ctypedef unsigned short WORD
    ctypedef float FLOAT
    ctypedef double DOUBLE
    ctypedef char CHAR
    ctypedef unsigned char UCHAR
    ctypedef void VOID
    ctypedef short SHORT
    ctypedef unsigned short USHORT
    ctypedef long long LONGLONG
    ctypedef char* LPSTR
    ctypedef unsigned long long ULONGLONG
    ctypedef long LONG
    ctypedef void* PVOID
    ctypedef DWORD* LPDWORD
    ctypedef int BOOL
    ctypedef const char* LPCTSTR
    ctypedef const wchar_t* LPWSTR
    ctypedef wchar_t* LPCWSTR
    ctypedef unsigned long long ULONG64
    ctypedef wchar_t* WCHAR
    ctypedef WCHAR* LPTSTR
    ctypedef wchar_t* LPSIDSTR
    ctypedef LPWSTR LPOLESTR
    ctypedef long HRESULT

    enum: ERROR_SUCCESS
    enum: ERROR_CANCELLED

    ctypedef struct GUID:
        DWORD Data1
        WORD Data2
        WORD Data3
        BYTE Data4[8]

    ctypedef const GUID & REFGUID

    ctypedef struct U:
        DWORD LowPart
        LONG HighPart
    ctypedef union LARGE_INTEGER:
        DWORD low"LowPart"
        LONG high "HighPart"
        U u
        LONGLONG QuadPart

    ctypedef struct FILETIME:
        DWORD low_date "dwLowDateTime"
        DWORD high_date "dwHighDateTime"

    ctypedef struct SYSTEMTIME:
        WORD year "wYear"
        WORD month "wMonth"
        WORD day_of_week "wDayOfWeek"
        WORD day "wDay"
        WORD hour "wHour"
        WORD minute "wMinute"
        WORD second "wSecond"
        WORD millis "wMilliseconds"

    ctypedef struct TIME_ZONE_INFORMATION:
        LONG       Bias
        WCHAR      StandardName[32]
        SYSTEMTIME StandardDate
        LONG       StandardBias
        WCHAR      DaylightName[32]
        SYSTEMTIME DaylightDate
        LONG       DaylightBias

    int string_from_guid "StringFromGUID2"(REFGUID guid, LPOLESTR lpsz, int cch) nogil

    BOOL filetime_to_systemtime "FileTimeToSystemTime"(FILETIME *ft, SYSTEMTIME *st) nogil

    BOOL systemtime_to_tz_specific_localtime "SystemTimeToTzSpecificLocalTime"(TIME_ZONE_INFORMATION *zone,
                                                                               SYSTEMTIME *uni_time,
                                                                               SYSTEMTIME *local_time) nogil

cdef extern from "winsock.h":
    USHORT ntohs(USHORT netshort) nogil
    ULONG  htonl(ULONG hostlong) nogil

    ctypedef union S_un:
        ULONG S_addr
    ctypedef struct in_addr:
        S_un S_un

    char* inet_ntoa(in_addr addr) nogil




# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from _ctypes import sizeof
from ctypes import byref, cast
from ctypes.wintypes import MAX_PATH
import os

from fibratus.kevent_types import CREATE_PROCESS, ENUM_PROCESS, TERMINATE_THREAD, TERMINATE_PROCESS, \
    CREATE_THREAD, ENUM_THREAD
from fibratus.apidefs.process import *
from fibratus.apidefs.cdefs import STATUS_SUCCESS
from fibratus.apidefs.sys import close_handle, malloc, free
from fibratus.common import DotD as ddict, NA


class ThreadRegistry(object):

    def __init__(self, handle_repository, handles, image_meta_registry):
        self._threads = {}
        self.on_thread_added_callback = None
        self.handle_repository = handle_repository
        self.image_meta_registry = image_meta_registry
        self._handles = handles

    def add_thread(self, ketype, kti):
        """Adds a new process or thread to thread registry.

        Parameters
        ----------

        ketype: tuple
            kernel event type
        kti: dict
            event payload as coming from the
            kernel event stream collector
        """
        if ketype == CREATE_PROCESS or ketype == ENUM_PROCESS:
            parent_pid = int(kti.parent_id, 16)
            process_id = int(kti.process_id, 16)
            # we assume the process id is
            # equal to thread id (in a single
            # threaded process)
            thread_id = process_id
            name = kti.image_file_name
            comm = kti.command_line

            thread = ThreadInfo(process_id, thread_id,
                                parent_pid,
                                name,
                                comm)
            if ketype == ENUM_PROCESS:
                thread.handles = [handle for handle in self._handles if handle.pid == process_id]
            if ketype == CREATE_PROCESS:
                image_meta = self.image_meta_registry.get_image_meta(thread.exe)
                if not image_meta:
                    image_meta = self.image_meta_registry.add_image_meta(thread.exe)
                thread.image_meta = image_meta
            self._threads[process_id] = thread

        elif ketype == CREATE_THREAD or ketype == ENUM_THREAD:
            # new thread created in the
            # context of the existing process
            # `procces_id` is the parent
            # of this thread
            process_id = int(kti.process_id, 16)
            parent_pid = process_id
            thread_id = int(kti.t_thread_id, 16)

            if parent_pid in self._threads:
                # copy info from the process
                # which created this thread
                pthread = self._threads[parent_pid]
                # increment the number of threads
                # for this process
                pthread.increment_child_count()

                name = pthread.name
                comm = pthread.comm

                thread = ThreadInfo(process_id, thread_id,
                                    parent_pid,
                                    name,
                                    comm)
                thread.ustack_base = hex(kti.user_stack_base)
                thread.kstack_base = hex(kti.stack_base)
                thread.base_priority = kti.base_priority
                thread.io_priority = kti.io_priority
                self._threads[thread_id] = thread
            else:
                # the parent process has not been found
                # query the os for process information
                handle = open_process(PROCESS_QUERY_INFORMATION | PROCESS_QUERY_LIMITED_INFORMATION | PROCESS_VM_READ,
                                      False,
                                      parent_pid)
                info = {}
                if handle:
                    info = self._query_process_info(handle)
                    close_handle(handle)
                else:
                    if get_last_error() == ERROR_ACCESS_DENIED:
                        if parent_pid == 0:
                            info = ddict(name='idle',
                                         comm='idle',
                                         parent_id=0)
                        else:
                            # the access to protected / system process
                            # can't be done with PROCESS_VM_READ or PROCESS_QUERY_INFORMATION
                            # flags. Open the process again but with
                            # restricted access rights, so we can get the process image file name
                            handle = open_process(PROCESS_QUERY_LIMITED_INFORMATION,
                                                  False,
                                                  parent_pid)
                            if handle:
                                info = self._query_process_info(handle, False)
                                close_handle(handle)

                # add a new thread and the parent process
                # we just found to avoid continuous lookup
                name = info.name if len(info) > 0 and info.name else NA
                comm = info.comm if len(info) > 0 and info.comm else NA
                ppid = info.parent_pid if len(info) > 0 and info.parent_pid else NA

                thread = ThreadInfo(process_id, thread_id,
                                    process_id,
                                    name,
                                    comm)
                thread.ustack_base = hex(kti.user_stack_base)
                thread.kstack_base = hex(kti.stack_base)
                thread.base_priority = kti.base_priority
                thread.io_priority = kti.io_priority

                parent = ThreadInfo(process_id, process_id,
                                    ppid,
                                    name,
                                    comm)
                # enumerate parent handles
                parent.handles = self.handle_repository.query_handles(process_id)

                self._threads[thread_id] = thread
                self._threads[parent_pid] = parent

        if self.on_thread_added_callback and callable(self.on_thread_added_callback):
            self.on_thread_added_callback(thread)

    def remove_thread(self, ketype, kti):
        """Removes the thread or process from the registry.

        Parameters
        ----------

        ketype: tuple
            kernel event type
        kti: dict
            event payload as coming from the
            kernel event stream collector
        """
        if ketype == TERMINATE_THREAD:
            thread_id = int(kti.t_thread_id, 16)
            if thread_id in self._threads:
                # remove the thread and
                # decrement the child count of
                # the parent process
                if self._threads[thread_id].child_count == 0:
                    thread = self._threads.pop(thread_id)
                    if thread and thread.pid in self._threads:
                        parent = self._threads[thread.pid]
                        parent.decrement_child_count()
        elif ketype == TERMINATE_PROCESS:
            # the process has exited
            # remove all of its threads
            process_id = int(kti.process_id, 16)
            if process_id in self._threads:
                proc = self._threads.pop(process_id)
                if proc.child_count > 0:
                    self._threads = dict((k, v) for k, v in self._threads.items()
                                         if v.child_count == 0 and k != process_id)

    def init_thread_kevent(self, kevent, ketype, kti):
        """Initialize kernel event.

        Parameters
        ----------

        kevent: KEvent
            instance of `KEvent` class
        ketype: tuple
            kernel event type
        kti: dict
            kernel event payload
        """
        if ketype == CREATE_THREAD or ketype == TERMINATE_THREAD:
            tid = int(kti.t_thread_id, 16)
            thread = self.get_thread(tid)
            if thread:
                kevent.params = dict(pid=thread.pid,
                                     tid=tid,
                                     kstack_base=hex(kti.stack_base),
                                     ustack_base=hex(kti.user_stack_base),
                                     io_priority=kti.io_priority,
                                     base_priority=kti.base_priority)
                kevent.pid = thread.pid
                kevent.tid = tid
        else:
            pid = int(kti.process_id, 16)
            thread = self.get_thread(pid)
            if thread:
                kparams = dict(pid=pid,
                               name=thread.name,
                               comm=thread.comm,
                               exe=thread.exe,
                               ppid=thread.ppid)
                if thread.image_meta:
                    # include image meta info
                    image_meta = dict(arch=thread.image_meta.arch,
                                      timestamp=thread.image_meta.timestamp,
                                      num_sections=thread.image_meta.num_sections,
                                      sections=thread.image_meta.sections,
                                      imports=thread.image_meta.imports,
                                      org=thread.image_meta.org,
                                      description=thread.image_meta.description,
                                      version=thread.image_meta.version,
                                      internal_name=thread.image_meta.internal_name,
                                      copyright=thread.image_meta.copyright)
                    kparams['image_meta'] = image_meta
                kevent.params = kparams
                kevent.pid = thread.ppid

    def set_thread_added_callback(self, callback):
        self.on_thread_added_callback = callback

    def get_thread(self, tid):
        return self._threads[tid] if tid in self._threads else None

    @property
    def threads(self):
        return self._threads

    def _query_process_info(self, handle, read_peb=True):
        """Gets an extended proc info.

        Parameters
        -----------

        handle: HANDLE
            handle to process for which the info
            should be acquired
        read_peb: boolean
            true in case the process PEB should be read

        """
        pbi_buff = malloc(sizeof(PROCESS_BASIC_INFORMATION))
        status = zw_query_information_process(handle,
                                              PROCESS_BASIC_INFO,
                                              pbi_buff,
                                              sizeof(PROCESS_BASIC_INFORMATION),
                                              byref(ULONG()))

        info = {}

        if status == STATUS_SUCCESS:
            pbi = cast(pbi_buff, POINTER(PROCESS_BASIC_INFORMATION))
            ppid = pbi.contents.inherited_from_unique_process_id
            if read_peb:
                # read the PEB to get the process parameters.
                # Because the PEB structure resides
                # in the address space of another process
                # we must read the memory block in order
                # to access the structure's fields
                peb_addr = pbi.contents.peb_base_address
                peb_buff = read_process_memory(handle, peb_addr, sizeof(PEB))
                if peb_buff:
                    peb = cast(peb_buff, POINTER(PEB))
                    # read the RTL_USER_PROCESS_PARAMETERS struct
                    # which contains the command line and the image
                    # name of the process
                    pp = peb.contents.process_parameters
                    pp_buff = read_process_memory(handle,
                                                  pp,
                                                  sizeof(RTL_USER_PROCESS_PARAMETERS))
                    if pp_buff:
                        pp = cast(pp_buff, POINTER(RTL_USER_PROCESS_PARAMETERS))

                        comm = pp.contents.command_line.buffer
                        comm_len = pp.contents.command_line.length
                        exe = pp.contents.image_path_name.buffer
                        exe_len = pp.contents.image_path_name.length

                        # these memory reads are required
                        # to copy the command line and image name buffers
                        cb = read_process_memory(handle, comm, comm_len)
                        eb = read_process_memory(handle, exe, exe_len)

                        if cb and eb:
                            # cast the buffers to
                            # UNICODE strings
                            comm = cast(cb, c_wchar_p).value
                            exe = cast(eb, c_wchar_p).value

                            # the image name contains the full path
                            # split the string to get the exec name
                            name = exe[exe.rfind('\\') + 1:]
                            info = ddict(name=name,
                                         comm=comm,
                                         parent_pid=ppid)
                            free(cb)
                            free(eb)
                        free(pp_buff)

                    free(peb_buff)
            else:
                # query only the process image file name
                exe = ctypes.create_unicode_buffer(MAX_PATH)
                size = DWORD(MAX_PATH)
                name = None
                status = query_full_process_image_name(handle,
                                                       0,
                                                       exe,
                                                       byref(size))
                if status:
                    exe = exe.value
                    name = exe[exe.rfind('\\') + 1:]
                info = ddict(name=name if name else NA,
                             comm=exe if type(exe) is str else None,
                             parent_pid=ppid)
        if pbi_buff:
            free(pbi_buff)

        return info


class ThreadInfo(object):
    """Represents the state of thread or process.
    """
    def __init__(self, pid, tid, ppid, name, comm):
        """Creates an instance of `ThreadInfo` class.

        Parameters
        ----------

        pid: int
            process identifier
        tid: int
            thread identifier in the scope of
            an existing process
        ppid: int
            parent process identifier
        name: str
            process name (cmd.exe)
        comm: str
            the full command line of a process
            (C:\Windows\system32\cmd.exe /cdir /-C /W)

        Attributes
        ----------

        exe: str
            the full name of the executable
            (C:\Windows\system32\cmd.exe)
        args: list
            command line arguments for the process
            (/cdir, /-C, /W)
        child_count: int
            the number of threads for this process
        handles: list
            a list of handles which owns the process
        ustack_base: int
            the base address of the thread user-space stack
        kstack_base: int
            the base address of the thread kernel-space stack
        io_priority: int
            thread I/O priority
        base_priority: int
            thread CPU priority
        """
        self._pid = pid
        self._tid = tid
        self._ppid = ppid

        # get the executable from the
        # full file system path
        head, _ = os.path.split(comm[0:comm.find('exe')])
        self._exe = '%s\%s' % (head, name)
        self._exe = self._exe.replace("\"", '')

        if 'SystemRoot' in self._exe:
            sys_root = os.path.expandvars("%SystemRoot%")
            self._exe = self._exe.replace('%SystemRoot%', sys_root)\
                .replace('\\SystemRoot', sys_root)

        self._name = name.lower() if NA not in name else NA
        self._comm = comm
        # the command line arguments
        # are separated by blank space
        self._args = comm.split()[1:]
        self._child_count = 0
        self._handles = []

        self._ustack_base = 0x0
        self._kstack_base = 0x0
        self._io_priority = 0
        self._base_priority = 0

        self._image_meta = None

    @property
    def pid(self):
        return self._pid

    @property
    def ppid(self):
        return self._ppid

    @property
    def tid(self):
        return self._tid

    @property
    def exe(self):
        return self._exe

    @property
    def name(self):
        return self._name

    @property
    def comm(self):
        return self._comm

    @property
    def args(self):
        return self._args

    @property
    def child_count(self):
        return self._child_count

    @property
    def handles(self):
        return self._handles

    @handles.setter
    def handles(self, handles):
        self._handles = handles

    @property
    def ustack_base(self):
        return self._ustack_base

    @ustack_base.setter
    def ustack_base(self, ustack_base):
        self._ustack_base = ustack_base

    @property
    def kstack_base(self):
        return self._kstack_base

    @kstack_base.setter
    def kstack_base(self, kstack_base):
        self._kstack_base = kstack_base

    @property
    def io_priority(self):
        return self._io_priority

    @io_priority.setter
    def io_priority(self, io_priority):
        self._io_priority = io_priority

    @property
    def base_priority(self):
        return self._base_priority

    @base_priority.setter
    def base_priority(self, base_priority):
        self._base_priority = base_priority

    @property
    def image_meta(self):
        return self._image_meta

    @image_meta.setter
    def image_meta(self, image_meta):
        self._image_meta = image_meta

    def increment_child_count(self):
        self._child_count += 1

    def decrement_child_count(self):
        if self._child_count != 0:
            self._child_count -= 1



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from ctypes import c_uint
import os

import fibratus.apidefs.declarer as declarer


class TestDeclarer():

    def test_declare_function(self):
        get_current_process_id = declarer.declare(declarer.KERNEL, 'GetCurrentProcessId', [], c_uint)

        assert callable(get_current_process_id)
        assert get_current_process_id.restype == c_uint
        assert get_current_process_id.argtypes is None

        # check the function result
        pid = os.getpid()
        assert get_current_process_id() == pid


# Copyright 2017 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import Mock, patch, MagicMock

from fibratus.kevent import KEvent
from logbook import Logger
import pytest
import os

from fibratus.errors import BindingError
from fibratus.output.amqp import AmqpOutput
from fibratus.thread import ThreadInfo


@pytest.fixture(scope='module')
def outputs():
    return dict(amqp=Mock(spec_set=AmqpOutput))


class TestYaraBinding(object):

    def test_init(self, outputs):
        with patch.dict('sys.modules', **{
            'yara': MagicMock(),
        }):
            from fibratus.binding.yar import YaraBinding
            with patch('os.path.exists', return_value=True), \
                 patch('os.path.isdir', return_value=True), \
                 patch('glob.glob', return_value=['silent_banker.yar']), \
                 patch('yara.compile') as yara_compile_mock:
                    YaraBinding(outputs,
                                Mock(spec_set=Logger), output='amqp', path='C:\\yara-rules')
                    yara_compile_mock.assert_called_with(os.path.join('C:\\yara-rules', 'silent_banker.yar'))

    def test_init_invalid_path(self, outputs):
        with patch.dict('sys.modules', **{
            'yara': None,
        }):
            from fibratus.binding.yar import YaraBinding
            with patch('os.path.exists', return_value=False), \
                 patch('os.path.isdir', return_value=False):
                    with pytest.raises(BindingError) as e:
                        YaraBinding(outputs,
                                    Mock(spec_set=Logger), output='amqp', path='C:\\yara-rules-invalid')
                        assert 'C:\\yara-rules-invalid rules path does not exist' in str(e.value)

    def test_init_yara_python_not_installed(self, outputs):
        with patch.dict('sys.modules', **{
            'yara': None,
        }):
            from fibratus.binding.yar import YaraBinding
            with pytest.raises(BindingError) as e:
                YaraBinding(outputs,
                            Mock(spec_set=Logger), output='amqp', path='C:\\yara-rules')
                assert 'yara-python package is not installed' in str(e.value)

    def test_run(self, outputs):
        with patch.dict('sys.modules', **{
            'yara': MagicMock(),
        }):
            from fibratus.binding.yar import YaraBinding
            with patch('os.path.exists', return_value=True), \
                 patch('os.path.isdir', return_value=True), \
                 patch('glob.glob', return_value=['silent_banker.yar']), \
                 patch('yara.compile'):
                yara_binding = YaraBinding(outputs,
                                           Mock(spec_set=Logger), output='amqp', path='C:\\yara-rules')
                yara_binding.run(thread_info=Mock(spec_set=ThreadInfo), kevent=Mock(spec_set=KEvent))
                assert yara_binding._rules.match.called



# Copyright 2015 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import patch, Mock

import pytest
import os

from fibratus.apidefs.cdefs import ERROR_ACCESS_DENIED
from fibratus.apidefs.process import PROCESS_QUERY_INFORMATION, PROCESS_QUERY_LIMITED_INFORMATION, PROCESS_VM_READ
from fibratus.handle import HandleRepository, HandleInfo
from fibratus.image_meta import ImageMetaRegistry
from fibratus.kevent_types import CREATE_PROCESS, CREATE_THREAD, ENUM_PROCESS, ENUM_THREAD, TERMINATE_THREAD, \
    TERMINATE_PROCESS
from fibratus.common import DotD as dd
from fibratus.thread import ThreadRegistry, ThreadInfo


@pytest.fixture(scope='module')
def handle_repo_mock():
    handle_repo = Mock(spec_set=HandleRepository)
    handle_repo.query_handles.return_value = [HandleInfo(20, 18446738026501927904, 'FILE',
                                                         'C:\\Windows\\System32\\kernel32.dll',
                                                         0x2d8)]
    return handle_repo


@pytest.fixture(scope='module')
def image_meta_registry_mock():
    imeta_meta_registry_mock = Mock(spec_set=ImageMetaRegistry)
    imeta_meta_registry_mock.get_image_meta.return_value = None
    return imeta_meta_registry_mock


@pytest.fixture(scope='module')
def thread_registry(handle_repo_mock, image_meta_registry_mock):
    p1 = {"session_id": 0, "command_line": "C:\\Windows\\system32\\services.exe", "process_id": "0x1e4",
          "unique_process_key": 18446738026492816176, "exit_status": 259,
          "parent_id": "0x17c",
          "image_file_name": "services.exe",
          "directory_table_base": 4299976704,
          "user_sid": None}
    t1 = {"user_stack_base": 14483456, "io_priority": 2, "teb_base": 8796092874752, "process_id": "0x1e4",
          "stack_limit": 18446735827441688576, "t_thread_id": "0x238", "base_priority": 9,
          "win32_start_addr": 2009573488, "page_priority": 5,
          "stack_base": 18446735827441713152, "user_stack_limit": 14450688,
          "thread_flags": 0, "affinity": 15, "sub_process_tag": "0x0"}
    t2 = {"user_stack_base": 16580608, "io_priority": 2, "teb_base": 8796092669952, "process_id": "0x1e4",
          "stack_limit": 18446735827442425856, "t_thread_id": "0x254", "base_priority": 9,
          "win32_start_addr": 8791752742020, "page_priority": 5,
          "stack_base": 18446735827442450432, "user_stack_limit": 16547840,
          "thread_flags": 0, "affinity": 15, "sub_process_tag": "0x0"}

    thread_registry = ThreadRegistry(handle_repo_mock, [], image_meta_registry_mock)
    thread_registry.add_thread(ENUM_PROCESS, dd(p1))
    thread_registry.add_thread(ENUM_THREAD, dd(t1))
    thread_registry.add_thread(ENUM_THREAD, dd(t2))

    return thread_registry


class TestThreadRegistry:

    def test_init_thread_registry(self, thread_registry):

        assert len(thread_registry.threads) == 3
        proc = thread_registry.threads[int('0x1e4', 16)]
        assert proc
        assert isinstance(proc, ThreadInfo)
        assert proc.child_count == 2

    def test_enum_process(self, thread_registry, handle_repo_mock):

        kti = dd({"session_id": 0, "command_line": "C:\\Windows\\system32\\svchost.exe -k RPCSS",
                  "process_id": "0x2d8",
                  "unique_process_key": 18446738026496154416,
                  "exit_status": 259, "user_sid": None, "parent_id": "0x1e4",
                  "image_file_name": "svchost.exe", "directory_table_base": 3716534272})
        thread_registry.add_thread(ENUM_PROCESS, kti)
        process_id = int(kti.process_id, 16)

        handle_repo_mock.query_handles.assert_not_called()

        t = thread_registry.get_thread(process_id)
        assert t

        assert t.pid == process_id
        assert t.ppid == int(kti.parent_id, 16)
        assert t.name == kti.image_file_name
        assert t.exe == 'C:\\Windows\\system32\\svchost.exe'
        assert t.comm == kti.command_line
        assert len(t.args) > 0
        assert ['-k', 'RPCSS'] == t.args
        assert '-k' in t.args
        assert 'RPCSS' in t.args

    def test_create_process(self, thread_registry, handle_repo_mock):

        kti = dd({"session_id": 4294967295, "command_line": "\\SystemRoot\\System32\\smss.exe",
                  "process_id": "0xfc", "unique_process_key": 18446738026484345648, "exit_status": 259,
                  "user_sid": None, "parent_id": "0x4", "image_file_name": "smss.exe",
                  "directory_table_base": 4508921856})
        thread_registry.add_thread(CREATE_PROCESS, kti)
        process_id = int(kti.process_id, 16)

        t = thread_registry.get_thread(process_id)
        sys_root = os.path.expandvars("%SystemRoot%")
        assert t
        assert t.pid == process_id
        assert t.ppid == int(kti.parent_id, 16)
        assert t.name == kti.image_file_name
        assert t.exe == '%s\\System32\\smss.exe' % sys_root
        assert t.comm == kti.command_line
        assert len(t.args) == 0

    def test_create_thread(self, thread_registry):

        kti = dd({"user_stack_base": 18874368, "io_priority": 2, "teb_base": 8796092882944,
                 "process_id": "0x1e4", "stack_limit": 18446735827462836224,
                 "t_thread_id": "0x57c", "base_priority": 9, "win32_start_addr": 2009592544,
                 "page_priority": 5, "stack_base": 18446735827462860800,
                 "user_stack_limit": 18841600, "thread_flags": 0,
                 "affinity": 15, "sub_process_tag": "0x0"})
        thread_registry.add_thread(CREATE_THREAD, kti)
        thread_id = int(kti.t_thread_id, 16)
        t = thread_registry.get_thread(thread_id)
        assert t
        assert t.tid == thread_id
        assert t.pid == t.ppid == int(kti.process_id, 16)
        assert t.kstack_base == hex(kti.stack_base)
        assert t.ustack_base == hex(kti.user_stack_base)
        assert t.base_priority == kti.base_priority
        assert t.io_priority == kti.io_priority
        assert t.child_count == 0

    @patch('fibratus.thread.open_process', return_value=13)
    @patch('fibratus.thread.close_handle', return_value=None)
    def test_create_thread_registry_proc_lookup_failed(self, close_handle_mock, open_process_mock,
                                                       thread_registry, handle_repo_mock):

        kti = dd({"user_stack_base": 10289152, "io_priority": 2, "teb_base": 8796092874752,
                  "process_id": "0x330", "stack_limit": 18446735827443150848, "t_thread_id": "0x338",
                  "base_priority": 8, "win32_start_addr": 2009573488,
                  "page_priority": 5, "stack_base": 18446735827443175424, "user_stack_limit": 10256384,
                  "thread_flags": 0, "affinity": 15, "sub_process_tag": "0x0"})
        thread_id = int(kti.t_thread_id, 16)
        process_id = int(kti.process_id, 16)

        with patch('fibratus.thread.ThreadRegistry._query_process_info', return_value=dd(name='Dwm.exe',
                   comm='C:\\Windows\\system32\\Dwm.exe',
                   parent_pid=0x748)) as query_process_info_mock:
            thread_registry.add_thread(CREATE_THREAD, kti)
            open_process_mock.assert_called_with(PROCESS_QUERY_INFORMATION |
                                                 PROCESS_QUERY_LIMITED_INFORMATION | PROCESS_VM_READ,
                                                 False, process_id)
            query_process_info_mock.assert_called_with(13)
            close_handle_mock.assert_called_with(13)

        handle_repo_mock.query_handles.assert_called_with(process_id)

        proc = thread_registry.get_thread(process_id)
        assert proc
        assert proc.name == 'dwm.exe'
        t = thread_registry.get_thread(thread_id)
        assert t
        assert t.name == 'dwm.exe'


    @patch('fibratus.thread.open_process', side_effect=[None, 29])
    @patch('fibratus.thread.close_handle', return_value=None)
    @patch('fibratus.thread.get_last_error', return_value=ERROR_ACCESS_DENIED)
    def test_create_thread_registry_proc_lookup_failed_invalid_handle(self, get_last_error, close_handle_mock,
                                                                      open_process_mock,
                                                                      thread_registry, handle_repo_mock):

        kti = dd({"user_stack_base": 1835008, "io_priority": 2, "teb_base": 8796092878848,
                  "process_id": "0x4c8", "stack_limit": 18446735827446923264, "t_thread_id": "0x4dc",
                  "base_priority": 8, "win32_start_addr": 4286166856, "page_priority": 5,
                  "stack_base": 18446735827446964224,
                  "user_stack_limit": 1744896, "thread_flags": 0, "affinity": 15, "sub_process_tag": "0x0"})
        thread_id = int(kti.t_thread_id, 16)
        process_id = int(kti.process_id, 16)

        with patch('fibratus.thread.ThreadRegistry._query_process_info', return_value=dd(name='explorer.exe',
                   comm='C:\\Windows\\Explorer.EXE',
                   parent_pid=0x4c8)) as query_process_info_mock:
            thread_registry.add_thread(CREATE_THREAD, kti)
            assert get_last_error.call_count == 1
            open_process_mock.assert_called_with(PROCESS_QUERY_LIMITED_INFORMATION,
                                                 False, process_id)

            query_process_info_mock.assert_called_with(29, False)
            close_handle_mock.assert_called_with(29)

        handle_repo_mock.query_handles.assert_called_with(process_id)

        proc = thread_registry.get_thread(process_id)
        assert proc
        assert proc.name == 'explorer.exe'
        t = thread_registry.get_thread(thread_id)
        assert t
        assert t.name == 'explorer.exe'
        assert t.comm == 'C:\\Windows\\Explorer.EXE'

    def test_terminate_thread(self, thread_registry):

        kti = dd({"user_stack_base": 18874368, "io_priority": 2, "teb_base": 8796092882944,
                 "process_id": "0x1e4", "stack_limit": 18446735827462836224,
                 "t_thread_id": "0x57c", "base_priority": 9, "win32_start_addr": 2009592544,
                 "page_priority": 5, "stack_base": 18446735827462860800,
                 "user_stack_limit": 18841600, "thread_flags": 0,
                 "affinity": 15, "sub_process_tag": "0x0"})

        thread_id = int(kti.t_thread_id, 16)
        process_id = int(kti.process_id, 16)
        proc = thread_registry.get_thread(process_id)

        child_count = proc.child_count
        t = thread_registry.get_thread(thread_id)
        assert t
        thread_registry.remove_thread(TERMINATE_THREAD, kti)

        t = thread_registry.get_thread(thread_id)
        assert t is None
        assert proc.child_count == child_count - 1

    def test_terminate_process(self, thread_registry):
        kti = dd({"session_id": 0, "command_line": "C:\\Windows\\system32\\services.exe", "process_id": "0x1e4",
                  "unique_process_key": 18446738026492816176, "exit_status": 259,
                  "parent_id": "0x17c",
                  "image_file_name": "services.exe",
                  "directory_table_base": 4299976704,
                  "user_sid": None})
        process_id = int(kti.process_id, 16)
        proc = thread_registry.get_thread(process_id)
        assert proc
        thread_registry.remove_thread(TERMINATE_PROCESS, kti)
        proc = thread_registry.get_thread(process_id)
        assert proc is None


# Copyright 2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from .windows cimport *

cdef extern from "evntcons.h":

    enum: EVENT_HEADER_FLAG_32_BIT_HEADER
    enum: PROCESS_TRACE_MODE_EVENT_RECORD

    ctypedef struct EVENT_TRACE_PROPERTIES:
        pass
    ctypedef struct ETW_BUFFER_CONTEXT:
        UCHAR cpuid "ProcessorNumber"
        UCHAR  Alignment
        USHORT LoggerId

    ctypedef struct EVENT_DESCRIPTOR:
        USHORT Id
        UCHAR Version
        UCHAR Channel
        UCHAR Level
        UCHAR opcode "Opcode"
        USHORT Task
        ULONGLONG Keyword

    ctypedef struct EVENT_HEADER:
        USHORT Size
        USHORT HeaderType
        USHORT flags "Flags"
        USHORT EventProperty
        ULONG thread_id "ThreadId"
        ULONG process_id "ProcessId"
        LARGE_INTEGER timestamp "TimeStamp"
        GUID ProviderId
        EVENT_DESCRIPTOR descriptor "EventDescriptor"
        ULONG KernelTime
        ULONG UserTime
        ULONGLONG ProcessorTime
        GUID ActivityId

    ctypedef struct LINKAGE:
        USHORT Linkage
        USHORT Reserverd2
    ctypedef struct EVENT_HEADER_EXTENDED_DATA_ITEM:
        USHORT    Reserved1
        USHORT    ExtType
        LINKAGE   Linkage
        USHORT	  DataSize
        ULONGLONG DataPtr

    ctypedef struct EVENT_RECORD:
        EVENT_HEADER header "EventHeader"
        ETW_BUFFER_CONTEXT buffer_ctx "BufferContext"
        USHORT ExtendedDataCount
        USHORT UserDataLength
        EVENT_HEADER_EXTENDED_DATA_ITEM* ExtendedData
        PVOID UserData
        PVOID user_ctx "UserContext"

cdef extern from "evntrace.h":

    ctypedef VOID (__stdcall *PEVENT_RECORD_CALLBACK) (EVENT_RECORD* e)

    ctypedef ULONG64 TRACEHANDLE

    enum: INVALID_PROCESSTRACE_HANDLE
    enum: EVENT_TRACE_REAL_TIME_MODE

    ctypedef struct EVENT_TRACE_LOGFILE:
        LPTSTR LogFileName
        LPSTR logger_name "LoggerName"
        ULONG LogFileMode
        ULONG trace_mode "ProcessTraceMode"
        PEVENT_RECORD_CALLBACK callback "EventRecordCallback"
        PVOID context "Context"

    TRACEHANDLE open_trace "OpenTrace"(EVENT_TRACE_LOGFILE* logfile)

    ULONG close_trace "CloseTrace"(TRACEHANDLE handle)

    ULONG process_trace "ProcessTrace"(TRACEHANDLE* handle, ULONG count,
                                       FILETIME* start,
                                       FILETIME* end)


# Copyright 2015/2016 by Nedim Sabic (RabbitStack)
# All Rights Reserved.
# http://rabbitstack.github.io

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pytest
from fibratus.kevent_types import *


@pytest.mark.parametrize('name, expected_kevent',
                         [(KEvents.CREATE_PROCESS, CREATE_PROCESS),
                         (KEvents.TERMINATE_PROCESS, TERMINATE_PROCESS),
                         (KEvents.CREATE_THREAD, CREATE_THREAD),
                         (KEvents.TERMINATE_THREAD, TERMINATE_THREAD),
                         (KEvents.REG_SET_VALUE, REG_SET_VALUE),
                         (KEvents.REG_QUERY_KEY, REG_QUERY_KEY),
                         (KEvents.REG_OPEN_KEY, REG_OPEN_KEY),
                         (KEvents.REG_QUERY_VALUE, REG_QUERY_VALUE),
                         (KEvents.REG_DELETE_KEY, REG_DELETE_KEY),
                         (KEvents.REG_DELETE_VALUE, REG_DELETE_VALUE),
                         (KEvents.REG_CREATE_KEY, REG_CREATE_KEY),
                         (KEvents.SEND, [SEND_SOCKET_UDPV4, SEND_SOCKET_TCPV4])])
def test_kname_to_tuple(name, expected_kevent):
    assert kname_to_tuple(name) == expected_kevent


def test_kname_to_tuple_unknown_kevent():
    with pytest.raises(UnknownKeventTypeError):
        kname_to_tuple('Exec')


def test_ktuple_to_name():
    pass


# Copyright 2016 by Nedim Sabic (RabbitStack)
# http://rabbitstack.github.io
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from libc.stddef cimport wchar_t

cdef extern from "wchar.h":
    int wprintf(const wchar_t *, ...) nogil
    int printf( const char* format, ... ) nogil
    int sprintf (char* str, const char* format, ... ) nogil
    long strtol(const char* nptr,  char** endptr, int base) nogil
    long wcstol(const wchar_t* nptr, wchar_t** endptr, int base) nogil
    wchar_t* _wcslwr(wchar_t * s) nogil
    int wcscmp(const wchar_t* string1, const wchar_t* string2) nogil
    size_t wcslen (const wchar_t* wcs)

cdef extern from "<string>" namespace "std" nogil:

    cdef cppclass wstring "std::wstring":
        wstring() except +
        wstring(wchar_t *) except +
        wstring(wchar_t *, size_t) except +
        wstring(wstring&) except +

        const wchar_t* data()
        size_t size()

        int compare(wstring&)

